# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import functools
import inspect
import typing
import uuid
from collections.abc import Iterable
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable

import sqlalchemy as sa

from pydiverse.common.util import deep_map
from pydiverse.common.util.computation_tracing import (
    ComputationTraceRef,
    ComputationTracerProxy,
)
from pydiverse.common.util.hashing import stable_hash
from pydiverse.pipedag import (
    Blob,
    ConfigContext,
    RawSql,
    Table,
    Task,
    TaskGetItem,
)
from pydiverse.pipedag.container import attach_annotation
from pydiverse.pipedag.context import RunContext, TaskContext
from pydiverse.pipedag.context.context import CacheValidationMode
from pydiverse.pipedag.core import UnboundTask
from pydiverse.pipedag.errors import CacheError
from pydiverse.pipedag.materialize.cache import TaskCacheInfo, task_cache_key

if TYPE_CHECKING:
    from pydiverse.pipedag.core import Flow, Stage
    from pydiverse.pipedag.materialize.store import PipeDAGStore
    from pydiverse.pipedag.materialize.table_hook_base import TableHook


class AutoVersionSupport(Enum):
    NONE = 0
    LAZY = 1
    TRACE = 2


class UnboundMaterializingTask(UnboundTask):
    """A materializing task without any bound arguments.

    Instances of this class get initialized using the
    :py:func:`@materialize <pydiverse.pipedag.materialize>` decorator.
    By calling this object inside a ``with Flow(...)`` context,
    a :py:class:`~.MaterializingTask` gets created that binds the provided
    arguments to the task.

    >>> @materialize
    ... def some_task(arg):
    ...     ...
    ...
    >>> type(some_task)
    <class 'pydiverse.pipedag.materialize.core.UnboundMaterializingTask'>

    >>> with Flow() as f:
    ...     with Stage("stage"):
    ...         x = some_task(...)
    ...
    >>> type(x)
    <class 'pydiverse.pipedag.materialize.core.MaterializingTask'>

    When you call a `UnboundMaterializingTask` outside a flow definition context,
    then the original function (that was decorated with :py:func:`@materialize
    <pydiverse.pipedag.materialize>`) gets called. The only difference being, that
    any :py:class:`~.Table`, :py:class:`~.Blob` or :py:class:`~.RawSql` objects
    returned from the task get replaced with the object that they wrap:

    >>> import pandas as pd
    >>>
    >>> @materialize(input_type=pd.DataFrame)
    ... def double_it(df):
    ...     return Table(df * 2)
    ...
    >>> x = pd.DataFrame({"x": [0, 1, 2, 3]})
    >>> double_it(x)
       x
    0  0
    1  2
    2  4
    3  6


    """

    def __init__(
        self,
        fn: Callable,
        *,
        name: str | None = None,
        input_type: type | None = None,
        version: str | None = None,
        cache: Callable[..., Any] | None = None,
        lazy: bool = False,
        group_node_tag: str | None = None,
        nout: int = 1,
        add_input_source: bool = False,
        ordering_barrier: bool | dict[str, Any] = False,
        call_context: Callable[[], Any] | None = None,
    ):
        super().__init__(
            MaterializationWrapper(fn),
            name=name,
            nout=nout,
        )

        self._bound_task_type = MaterializingTask
        self._original_fn = fn

        self.input_type = input_type
        self.version = version
        self.cache = cache
        self.lazy = lazy
        self.group_node_tag = group_node_tag
        self.add_input_source = add_input_source
        self.call_context = call_context

        if isinstance(ordering_barrier, bool):
            group_node_args = {"ordering_barrier": ordering_barrier}
        elif isinstance(ordering_barrier, dict):
            group_node_args = ordering_barrier.copy()
            if "ordering_barrier" not in group_node_args:
                group_node_args["ordering_barrier"] = True

            group_node_keys = {"ordering_barrier", "label", "style", "style_tag"}
            unexpected_keys = set(ordering_barrier.keys()) - group_node_keys
            if unexpected_keys:
                raise ValueError(
                    "If ordering_barrier is a dictionary, it must only contain keys "
                    "that are valid parameters of class GroupNode constructor."
                    f"{unexpected_keys} are unexpected."
                )
        else:
            raise ValueError("The ordering_barrier parameter must be a boolean or a dictionary.")
        self.group_node_args = group_node_args

        if version is AUTO_VERSION:
            if input_type is None:
                # The input_type is used to determine which type of auto versioning
                # to use (either not supported, lazy or tracing)
                raise ValueError("Auto-versioning task must specify an input type")
            if lazy:
                raise ValueError("Task can't be lazy and auto-versioning at the same time")

    def __call__(self, *args, **kwargs) -> "MaterializingTask":
        if self.group_node_args["ordering_barrier"]:
            from pydiverse.pipedag import GroupNode

            with GroupNode(**self.group_node_args):
                return super().__call__(*args, **kwargs)  # type: ignore
        return super().__call__(*args, **kwargs)  # type: ignore

    def _call_original_function(self, *args, **kwargs):
        try:
            task_context = TaskContext.get()  # this may raise Lookup Error

            # this is a subtask call. Thus we demand identical input_types
            sub_task: MaterializingTask = task_context.task  # type: ignore
            if (
                sub_task.input_type is not None
                and self.input_type is not None
                and sub_task.input_type != self.input_type
            ):
                raise RuntimeError(
                    f"Subtask input type {self.input_type} does not match parent task "
                    f"input type {sub_task.input_type}: "
                    f"task={self}, sub_task={sub_task}"
                )
        except LookupError:
            # this is expected when calling the task outside of flow declaration
            pass
        result = self._original_fn(*args, **kwargs)

        def unwrap_mutator(x):
            if isinstance(x, (Table, Blob)):
                return x.obj
            if isinstance(x, RawSql):
                return x.sql
            return x

        return deep_map(result, unwrap_mutator)


class MaterializingTaskGetItem(TaskGetItem):
    """Object that represents a subset of a :py:class:`~.MaterializingTask` output.

    Instances of this class get initialized by calling
    :py:class:`MaterializingTask.__getitem__`.
    """

    def __init__(
        self,
        task: "MaterializingTask",
        parent: "MaterializingTask | MaterializingTaskGetItem",
        item: Any,
    ):
        super().__init__(task, parent, item)

    def __getitem__(self, item) -> "MaterializingTaskGetItem":
        """
        Same as :py:meth:`MaterializingTask.__getitem__`,
        except that it allows you to further refine the selection.
        """
        return super().__getitem__(item)

    def get_output_from_store(self, as_type: type = None, ignore_position_hashes: bool = False) -> Any:
        """
        Same as :py:meth:`MaterializingTask.get_output_from_store()`,
        except that it only loads the required subset of the output.
        """
        from pydiverse.pipedag.materialize.core import _get_output_from_store

        return _get_output_from_store(self, as_type, ignore_position_hashes=ignore_position_hashes)


class MaterializingTask(Task):
    """
    A pipedag task that materializes all its outputs.

    Instances of this class get initialized by calling a
    :py:class:`~.UnboundMaterializingTask`.

    As a user, you will encounter `MaterializingTask` objects, together with
    :py:class:`~.MaterializingTaskGetItem` objects, mostly during flow declaration,
    because they are used to bind the output (or parts of the output) from one
    task, to the input of another task.
    """

    def __init__(
        self,
        unbound_task: UnboundMaterializingTask,
        bound_args: inspect.BoundArguments,
        flow: "Flow",
        stage: "Stage",
    ):
        super().__init__(unbound_task, bound_args, flow, stage)

        self.input_type = unbound_task.input_type
        self.add_input_source = unbound_task.add_input_source
        self._version = unbound_task.version
        self.cache = unbound_task.cache
        self.lazy = unbound_task.lazy
        self.fn_annotations = typing.get_type_hints(unbound_task.fn)

    @property
    def version(self):
        try:
            if v := TaskContext.get().override_version:
                return v
        except LookupError:
            pass

        return self._version

    @version.setter
    def version(self, version):
        try:
            TaskContext.get().override_version = version
        except LookupError:
            raise AttributeError("Can't set task version outside of a TaskContext") from None

    def __getitem__(self, item) -> MaterializingTaskGetItem:
        """Construct a :py:class:`~.MaterializingTaskGetItem`.

        If the corresponding task returns an object that supports
        :external+python:py:meth:`__getitem__ <object.__getitem__>`,
        then this allows you to forward only parts of the task's output to
        the next task.

        Example
        -------
        ::

            @materialize
            def dict_task():
                return {"x": 1, "y": 2}

            @materialize
            def assert_equal(actual, expected):
                assert actual == expected

            with Flow():
                with Stage("stage"):
                    d = dict_task()

                    # Only pass parts of the dictionary returned by dict_task
                    # to the assert_equal tasks.
                    assert_equal(d["x"], 1)
                    assert_equal(d["y"], 2)

        """
        return MaterializingTaskGetItem(self, self, item)

    def run(self, inputs: dict[int, Any], ignore_position_hashes: bool = False, **kwargs):
        # When running only a subset of an entire flow, not all inputs
        # get calculated during flow execution. As a consequence, we must load
        # those inputs from the cache.

        def replace_stage_with_pseudo_stage(x):
            if isinstance(x, (Table, Blob, RawSql)):
                from pydiverse.pipedag.materialize.core import PseudoStage

                x.stage = PseudoStage(x.stage, did_commit=True)
            return x

        for in_id, in_task in self.input_tasks.items():
            if in_id in inputs:
                continue

            store = ConfigContext.get().store
            cached_output, metadata = store.retrieve_most_recent_task_output_from_cache(in_task, ignore_position_hashes)

            cached_output = deep_map(cached_output, replace_stage_with_pseudo_stage)
            inputs[in_id] = cached_output

        return super().run(inputs, **kwargs)

    def get_output_from_store(self, as_type: type = None, ignore_position_hashes: bool = False) -> Any:
        """Retrieves the output of the task from the cache.

        No guarantees are made regarding whether the returned values are still
        up-to-date and cache valid.

        :param as_type: The type as which tables produced by this task should
            be dematerialized. If no type is specified, the input type of
            the task is used.
        :param ignore_position_hashes:
            If ``True``, the position hashes of tasks are not checked
            when retrieving the inputs of a task from the cache.
            This simplifies execution of subgraphs if you don't care whether inputs to
            that subgraph are cache invalid. This allows multiple modifications in the
            Graph before the next run updating the cache.
            Attention: This may break automatic cache invalidation.
            And for this to work, any task producing an input
            for the chosen subgraph may never be used more
            than once per stage.
        :return: The output of the task.
        :raise CacheError: if no outputs for this task could be found in the store.

        Example
        -------

        ::

            # Define some flow
            with Flow() as f:
                with Stage("stage_1"):
                    x = some_task()
                    ...

            # Get output BEFORE calling flow.run()
            df_x = x.get_output_from_store(as_type=pd.DataFrame)

        """
        from pydiverse.pipedag.materialize.core import _get_output_from_store

        return _get_output_from_store(self, as_type, ignore_position_hashes=ignore_position_hashes)


class AutoVersionType:
    """
    Special constant used to indicate that pipedag should automatically determine
    a version number for a task.

    The version is determined by running the task once to construct a representation
    of the computation performed on the tables (e.g. construct a query plan /
    computational graph).
    Using this representation, a unique version number is constructed such that
    if the computation changes the version number also changes.
    Then, if the task is deemed to be cache invalid, it is run again, but this
    time with actual data.

    This puts the following constraints on which tasks can use `AUTO_VERSION`:

    * The task must be a pure function.
    * The task may not inspect the contents of the input tables.
    * The task must return at least one table.
    * The task may not return :py:class:`RawSql`.
    * The task may not return :py:class:`Blob`.

    .. rubric:: Polars

    Automatic versioning is best supported by polars.
    To use it, you must specify
    :external+pl:doc:`polars.LazyFrame <reference/lazyframe/index>`
    as the task input type and only use LazyFrames inside your task.

    .. rubric:: Pandas

    Pandas support is still *experimental*. It is implemented using a technique
    we call "computation tracing", where we run the task with proxy tables
    as inputs that record all operations performed on them.
    This requires heavily monkey patching the pandas and numpy modules.
    As long as you only use pandas and numpy functions inside your task,
    computation tracing should work successfully.
    However, for the monkey patching to work, you only access pandas / numpy
    functions through their namespace (e.g. ``pd.concat(...)`` is allowed, while
    ``from pandas import concat; concat(...)`` is not allowed).

    Requires dask to be installed.


    Example
    -------
    ::

        @materialize(input_type=pl.LazyFrame, version=AUTO_VERSION)
        def polars_task(x: pl.LazyFrame):
            # Some operations that only utilize pl.LazyFrame...
            return x.with_columns(
                (pl.col("col1") * pl.col("col2")).alias("col3")
            )

    """

    def __str__(self):
        return "AUTO_VERSION"

    def __repr__(self):
        return "AUTO_VERSION"


AUTO_VERSION = AutoVersionType()


class MaterializationWrapper:
    """Function wrapper that contains all high level materialization logic

    :param fn: The function to wrap
    """

    def __init__(self, fn: Callable):
        functools.update_wrapper(self, fn)

        self.fn = fn
        self.fn_signature = inspect.signature(fn)
        self.fn_annotations = typing.get_type_hints(fn)

    def __call__(self, *args, **kwargs):
        """Function wrapper / materialization logic

        :param args: The arguments passed to the function
        :param _pipedag_task_: The `MaterializingTask` instance which called
            this wrapper.
        :param kwargs: The keyword arguments passed to the function
        :return: A copy of what the original function returns annotated
            with some additional metadata.
        """

        task_context = TaskContext.get()
        config_context = ConfigContext.get()
        run_context = RunContext.get()

        task: MaterializingTask = task_context.task  # type: ignore
        bound = self.fn_signature.bind(*args, **kwargs)

        for name, value in bound.arguments.items():
            if name in self.fn_annotations:
                # copy type annotation into Table objects so they can be used during
                # dematerialization
                attach_annotation(self.fn_annotations[name], value)

        if task is None:
            raise TypeError("Task can't be None.")

        # If this is the first task in this stage to be executed, ensure that
        # the stage has been initialized and locked.
        store = config_context.store
        store.ensure_stage_is_ready(task.stage)

        # Compute the cache key for the task inputs
        input_json = store.json_encode(bound.arguments)
        input_hash = stable_hash("INPUT", input_json)

        cache_fn_hash = ""
        if task.cache is not None:
            if config_context.cache_validation.disable_cache_function:
                # choose random cache_fn_hash to ensure downstream tasks are
                # invalidated for FORCE_FRESH_INPUT
                cache_fn_hash = stable_hash("CACHE_FN", uuid.uuid4().hex)
            else:
                cache_fn_output = store.json_encode(task.cache(*args, **kwargs))
                cache_fn_hash = stable_hash("CACHE_FN", cache_fn_output)

        memo_cache_key = task_cache_key(task, input_hash, cache_fn_hash)

        # Check if this task has already been run with the same inputs
        # If yes, return memoized result. This prevents DuplicateNameExceptions
        with run_context.task_memo(task, memo_cache_key) as (success, memo):
            if success:
                task.logger.info("Task has already been run with the same inputs. Using memoized results.")

                return memo

            # Cache Lookup (only required if task isn't lazy)
            force_task_execution = config_context.cache_validation.mode == CacheValidationMode.FORCE_CACHE_INVALID or (
                config_context.cache_validation.mode == CacheValidationMode.FORCE_FRESH_INPUT and task.cache is not None
            )
            skip_cache_lookup = (
                config_context.cache_validation.ignore_task_version and task.version != AUTO_VERSION
            ) or task.version is None
            assert_no_fresh_input = config_context.cache_validation.mode == CacheValidationMode.ASSERT_NO_FRESH_INPUT

            if task.version is AUTO_VERSION:
                if force_task_execution and config_context.cache_validation.disable_cache_function:
                    task.version = None
                    skip_cache_lookup = True
                    task.logger.info(
                        "Skipping determination of auto version of task due to forced "
                        "execution and disable_cache_function=True"
                    )
                else:
                    assert not task.lazy

                    auto_version = self.compute_auto_version(task, store, bound)
                    task.version = auto_version
                    task.logger.info("Assigned auto version to task", version=auto_version)

            if not task.lazy and not skip_cache_lookup and not force_task_execution:
                cache_metadata = None
                try:
                    cached_output, cache_metadata = store.retrieve_cached_output(task, input_hash, cache_fn_hash)
                except CacheError as e:
                    task.logger.info(
                        "Failed to retrieve task from cache",
                        input_hash=input_hash,
                        cache_fn_hash=cache_fn_hash,
                        version=task.version,
                        cause=str(e),
                    )
                    TaskContext.get().is_cache_valid = False
                    run_context.trace_hook.task_cache_status(task, input_hash, cache_fn_hash, cache_valid=False)

                if cache_metadata is not None:
                    store.copy_cached_output_to_transaction_stage(cached_output, cache_metadata, task)
                    run_context.store_task_memo(task, memo_cache_key, cached_output)
                    task.logger.info(
                        "Found task in cache. Using cached result.",
                        input_hash=input_hash,
                        cache_fn_hash=cache_fn_hash,
                    )
                    TaskContext.get().is_cache_valid = True
                    run_context.trace_hook.task_cache_status(
                        task, input_hash, cache_fn_hash, cache_metadata, cached_output
                    )
                    return cached_output

            if not task.lazy:
                if assert_no_fresh_input and task.cache is not None:
                    raise AssertionError(
                        "cache_validation.mode=ASSERT_NO_FRESH_INPUT is a "
                        "protection mechanism to prevent execution of "
                        "source tasks to keep pipeline input stable. However,"
                        "this task was still triggered."
                    ) from None

            if task.lazy:
                # For lazy tasks, is_cache_valid gets set to false during the
                # store.materialize_task procedure
                TaskContext.get().is_cache_valid = True

                if config_context.cache_validation.mode == CacheValidationMode.IGNORE_FRESH_INPUT:
                    # `cache_fn_hash` is not used for cache retrieval if
                    # ignore_cache_function is set to True.
                    # In that case, cache_metadata.cache_fn_hash may be different
                    # from the cache_fn_hash of the current task run.

                    try:
                        _, cache_metadata = store.retrieve_cached_output(task, input_hash, "")
                        cache_fn_hash = cache_metadata.cache_fn_hash
                        run_context.trace_hook.task_cache_status(
                            task, input_hash, cache_fn_hash, cache_metadata, lazy=True
                        )
                    except CacheError as e:
                        task.logger.info(
                            "Failed to retrieve lazy task from cache",
                            cause=str(e),
                            input_hash=input_hash,
                            version=task.version,
                            cache_fn_hash=cache_fn_hash,
                        )

            # Prepare TaskCacheInfo
            if not task.lazy and skip_cache_lookup:
                # Assign random version to disable caching for this task
                task = copy.copy(task)
                task.version = uuid.uuid4().hex

            task_cache_info = TaskCacheInfo(
                task=task,
                input_hash=input_hash,
                cache_fn_hash=cache_fn_hash,
                cache_key=task_cache_key(task, input_hash, cache_fn_hash),
                assert_no_materialization=assert_no_fresh_input,
                force_task_execution=force_task_execution,
            )

            # Compute the input_tables value of the TaskContext
            input_tables = []

            def _input_tables_visitor(x):
                if isinstance(x, Table):
                    input_tables.append(x)
                return x

            deep_map(bound.arguments, _input_tables_visitor)
            task_context.input_tables = input_tables

            # Not found in cache / lazy -> Evaluate Function
            args, kwargs = store.dematerialize_task_inputs(task, bound.args, bound.kwargs)

            def imperative_materialize(
                table: Table,
                config_context: ConfigContext | None,
                return_as_type: type | None = None,
                return_nothing: bool = False,
            ):
                my_store = config_context.store if config_context is not None else store
                state = task_cache_info.imperative_materialization_state
                if id(table) in state.table_ids:
                    raise RuntimeError("The table has already been imperatively materialized.")
                table.assumed_dependencies = (
                    list(sorted(state.assumed_dependencies)) if len(state.assumed_dependencies) > 0 else []
                )
                _ = my_store.materialize_task(task, task_cache_info, table, disable_task_finalization=True)
                if not return_nothing:

                    def get_return_obj(return_as_type):
                        if return_as_type is None:
                            return_as_type = task.input_type
                            if return_as_type is None or not my_store.table_store.get_r_table_hook(
                                return_as_type
                            ).retrieve_as_reference(return_as_type):
                                # dematerialize as sa.Table if it would transfer all
                                # rows to python when dematerializing with input_type
                                return_as_type = sa.Table
                        obj = my_store.dematerialize_item(table, return_as_type, run_context)
                        state.add_table_lookup(obj, table)
                        return obj

                    if isinstance(return_as_type, Iterable):
                        return tuple(get_return_obj(t) for t in return_as_type)
                    else:
                        return get_return_obj(return_as_type)

            task_context.imperative_materialize_callback = imperative_materialize
            run_context.trace_hook.task_pre_call(task)
            result = self.fn(*args, **kwargs)
            run_context.trace_hook.task_post_call(task)
            task_context.imperative_materialize_callback = None
            if task.debug_tainted:
                raise RuntimeError(f"The task {task.name} has been tainted by interactive debugging. Aborting.")

            def result_finalization_mutator(x):
                state = task_cache_info.imperative_materialization_state
                object_lookup = state.object_lookup
                if id(x) in object_lookup:
                    # substitute imperatively materialized object references with
                    # their respective table objects
                    x = object_lookup[id(x)]
                if isinstance(x, (Table, RawSql)):
                    # fill assumed_dependencies for Tables that were not yet
                    # materialized
                    if len(state.assumed_dependencies) > 0:
                        if x.assumed_dependencies is None:
                            x.assumed_dependencies = list(sorted(state.assumed_dependencies))
                return x

            result = deep_map(result, result_finalization_mutator)
            result = store.materialize_task(task, task_cache_info, result)

            # Delete underlying objects from result (after materializing them)
            def obj_del_mutator(x):
                if isinstance(x, (Table, Blob)):
                    x.obj = None
                if isinstance(x, RawSql):
                    x.loaded_tables = None
                if isinstance(x, (Table, RawSql)):
                    x.assumed_dependencies = deep_map(x.assumed_dependencies, obj_del_mutator)
                return x

            result = deep_map(result, obj_del_mutator)
            run_context.store_task_memo(task, memo_cache_key, result)
            self.value = result
            run_context.trace_hook.task_complete(task, result)

            return result

    def compute_auto_version(
        self,
        task: MaterializingTask,
        store: "PipeDAGStore",
        bound: inspect.BoundArguments,
    ) -> str:
        # Decide on which type of auto versioning to use.
        # For this, we go off the `auto_version_support` flag of the retrieve
        # table hook of the input type

        hook = store.table_store.get_r_table_hook(task.input_type)
        auto_version_support = hook.auto_version_support

        if auto_version_support == AutoVersionSupport.NONE:
            raise TypeError(f"Auto versioning not supported for input type {task.input_type}.")

        # Perform auto versioning
        info = None
        if auto_version_support == AutoVersionSupport.LAZY:
            info = self.compute_auto_version_lazy(task, hook, store, bound)
        elif auto_version_support == AutoVersionSupport.TRACE:
            info = self.compute_auto_version_trace(task, hook, store, bound)

        assert isinstance(info, str)

        auto_version = "auto_" + stable_hash("AUTO-VERSION", info)
        return auto_version

    def compute_auto_version_lazy(
        self,
        task: MaterializingTask,
        hook: type["TableHook"],
        store: "PipeDAGStore",
        bound: inspect.BoundArguments,
    ) -> str:
        args, kwargs = store.dematerialize_task_inputs(task, bound.args, bound.kwargs, for_auto_versioning=True)

        try:
            result = self.fn(*args, **kwargs)
        except Exception as e:
            raise RuntimeError("Auto versioning failed") from e

        result, tables = self._auto_version_prepare_and_check_task_output(task, store, result)

        if len(tables) == 0:
            raise ValueError(f"Task with version={AUTO_VERSION} must return at least one Table.")

        # Compute auto version
        auto_version_info = []
        for table in tables:
            obj = table.obj
            auto_version_info.append(hook.get_auto_version_lazy(obj))

        return store.json_encode(
            {
                "task_output": result,
                "auto_version_info": auto_version_info,
            }
        )

    def compute_auto_version_trace(
        self,
        task: MaterializingTask,
        hook: type["TableHook"],
        store: "PipeDAGStore",
        bound: inspect.BoundArguments,
    ) -> str | None:
        computation_tracer = hook.get_computation_tracer()

        def tracer_proxy_mapper(x):
            if isinstance(x, Table):
                return computation_tracer.create_proxy(("TABLE", x.stage.name, x.name))
            if isinstance(x, RawSql):
                return computation_tracer.create_proxy(("RAW_SQL", x.stage.name, x.name, x.cache_key))
            if isinstance(x, Blob):
                return computation_tracer.create_proxy(("BLOB", x.stage.name, x.name))

            return x

        args = deep_map(bound.args, tracer_proxy_mapper)
        kwargs = deep_map(bound.kwargs, tracer_proxy_mapper)

        with computation_tracer:
            try:
                result = self.fn(*args, **kwargs)
            except Exception as e:
                raise RuntimeError("Auto versioning failed") from e

        result, tables = self._auto_version_prepare_and_check_task_output(task, store, result)

        # Compute auto version
        trace_hash = computation_tracer.trace_hash()
        table_info = []

        # Include the computation tracers in the returned tables in the version
        # to invalidate the task when outputs get switched
        for table in tables:
            obj = table.obj
            assert isinstance(obj, ComputationTracerProxy)
            table_info.append(ComputationTraceRef(obj).id)

        return store.json_encode(
            {
                "task_output": result,
                "table_info": table_info,
                "trace_hash": trace_hash,
            }
        )

    def _auto_version_prepare_and_check_task_output(self, task: MaterializingTask, store: "PipeDAGStore", output):
        task_cache_info = TaskCacheInfo(
            task=task,
            input_hash="AUTO_VERSION",
            cache_fn_hash="AUTO_VERSION",
            cache_key="AUTO_VERSION",
            assert_no_materialization=False,
            force_task_execution=False,
        )

        # Replace computation tracer proxy objects in output
        def replace_proxy(x):
            if isinstance(x, ComputationTracerProxy):
                return {
                    "__pipedag_type__": "computation_tracer_proxy",
                    "id": ComputationTraceRef(x).id,
                }
            return x

        output = deep_map(output, replace_proxy)
        output, tables, raw_sqls, blobs = store.prepare_task_output_for_materialization(task, task_cache_info, output)

        if len(raw_sqls) != 0:
            raise ValueError(f"Task with version={AUTO_VERSION} can't return RawSql.")

        if len(blobs) != 0:
            raise ValueError(f"Task with version={AUTO_VERSION} can't return Blobs.")

        return output, tables



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import datetime
import importlib.metadata
import inspect
import os
import sys

import sphinx.util
from sphinx.ext.napoleon import GoogleDocstring


sys.path.insert(0, os.path.abspath("../.."))
sys.path.insert(0, os.path.abspath("../../src"))
os.environ["SPHINX_BUILD"] = "1"

_mod = importlib.import_module("pydiverse.pipedag")

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "pydiverse.pipedag"
copyright = f"{datetime.date.today().year}, QuantCo, Inc"
author = "QuantCo, Inc."

release = importlib.metadata.version("pydiverse.pipedag")

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.linkcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
    "sphinx.ext.napoleon",
    "sphinx_click",
]

myst_enable_extensions = [
    "fieldlist",
    "deflist",
    "attrs_block",
    "attrs_inline",
]

autodoc_default_options = {
    "member-order": "bysource",
}

autosectionlabel_prefix_document = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "sa": ("https://docs.sqlalchemy.org/en/stable", None),
    "pd": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "pl": ("https://pola-rs.github.io/polars/py-polars/html", None),
    "tp": ("https://tidypolars.readthedocs.io/en/stable", None),
    # "ibis": ("https://ibis-project.org/", None),
    "kazoo": ("https://kazoo.readthedocs.io/en/latest", None),
    "dask": ("https://docs.dask.org/en/stable", None),
}

napoleon_custom_sections = [
    "Config File",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
# html_static_path = ["_static"]
html_title = "pydiverse.pipedag"

html_theme_options = {
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/pydiverse/pydiverse.pipedag/",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
    "source_repository": "https://github.com/pydiverse/pydiverse.pipedag/",
    "source_branch": "main",
    "source_directory": "docs/",
}


# Fix signature
# Use __init__ instead of __new__
def _correct_signatures(app, what, name, obj, options, signature, return_annotation):
    if "__overload__" in str(signature):
        return str(signature).replace("__overload__", ""), return_annotation

    if what == "class":
        if obj.__init__ != object.__init__:
            signature = sphinx.util.inspect.signature(obj.__init__)

            parameters = signature.parameters.copy()
            parameters.popitem(last=False)  # Pop first to remove self

            signature = signature.replace(parameters=parameters.values())
            signature = str(signature)

    return signature, return_annotation


def setup(app):
    app.connect("autodoc-process-signature", _correct_signatures)
    pass


# Copied and adapted from
# https://github.com/pandas-dev/pandas/blob/4a14d064187367cacab3ff4652a12a0e45d0711b/doc/source/conf.py#L613-L659
# Required configuration function to use sphinx.ext.linkcode
def linkcode_resolve(domain, info):
    """Determine the URL corresponding to a given Python object."""
    try:
        if domain != "py":
            return None

        module_name = info["module"]
        full_name = info["fullname"]

        _submodule = sys.modules.get(module_name)
        if _submodule is None:
            return None

        _object = _submodule
        for _part in full_name.split("."):
            try:
                _object = getattr(_object, _part)
            except AttributeError:
                return None

        try:
            fn = inspect.getsourcefile(inspect.unwrap(_object))  # type: ignore
        except TypeError:
            fn = None
        if not fn:
            return None

        try:
            source, line_number = inspect.getsourcelines(_object)
        except OSError:
            line_number = None  # type: ignore

        if line_number:
            linespec = f"#L{line_number}-L{line_number + len(source) - 1}"
        else:
            linespec = ""

        fn = os.path.relpath(fn, start=os.path.dirname(str(_mod.__file__)))

        return (
            "https://github.com/pydiverse/pydiverse.pipedag"
            f"/tree/{release}/src/pydiverse/pipedag/{fn}{linespec}"
        )
    except Exception as e:
        return None


# Patch napoleon so that the type in the attributes section gets placed on the
# same line as the name of the attribute
def _parse_attributes_section(self, section: str) -> list[str]:
    lines = []
    for _name, _type, _desc in self._consume_fields():
        if not _type:
            _type = self._lookup_annotation(_name)
        if self._config.napoleon_use_ivar:
            field = ":ivar %s: " % _name
            lines.extend(self._format_block(field, _desc))
            if _type:
                lines.append(f":vartype {_name}: {_type}")
        else:
            lines.append(".. attribute:: " + _name)
            if _type:
                lines.extend(self._indent([":type: %s" % _type], 3))
            if self._opt and "noindex" in self._opt:
                lines.append("   :noindex:")
            lines.append("")

            fields = self._format_field("", "", _desc)
            lines.extend(self._indent(fields, 3))
            lines.append("")
    if self._config.napoleon_use_ivar:
        lines.append("")
    return lines


GoogleDocstring._parse_attributes_section = _parse_attributes_section



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import warnings
from typing import Any

import structlog
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from pydiverse.common.util import requires
from pydiverse.pipedag import ExternalTableReference, Table
from pydiverse.pipedag.context import ConfigContext, RunContext
from pydiverse.pipedag.core import Subflow, Task
from pydiverse.pipedag.core.result import Result
from pydiverse.pipedag.engine.base import (
    OrchestrationEngine,
)

try:
    import prefect

    prefect_version = Version(prefect.__version__)
except ImportError as e:
    warnings.warn(str(e), ImportWarning)

    prefect = None
    prefect_version = Version("0")


@requires(prefect, ImportError("Module 'prefect' not installed"))
@requires(
    prefect_version in SpecifierSet("~=1.0"),
    ImportWarning(f"Requires prefect version 1.x (found {prefect_version})"),
)
class PrefectOneEngine(OrchestrationEngine):
    """
    Hands over execution of a flow to `Prefect 1 <https://docs-v1.prefect.io>`_.

    :param flow_kwargs:
        Optional dictionary of keyword arguments that get passed to the
        initializer of |prefect1.Flow|_.

    .. |prefect1.Flow| replace:: ``prefect.Flow``
    .. _prefect1.Flow: https://docs-v1.prefect.io/api/latest/core/flow.html
    """

    def __init__(self, flow_kwargs: dict[str, Any] = None):
        self.flow_kwargs = flow_kwargs or {}
        self.logger = structlog.get_logger(stage=self)

    def construct_prefect_flow(
        self,
        f: Subflow,
        inputs: dict[Task, ExternalTableReference] | None = None,
    ):
        inputs = inputs if inputs is not None else {}

        run_context = RunContext.get()
        config_context = ConfigContext.get()

        flow_kwargs = {
            "name": f.name,
            **self.flow_kwargs,
        }

        flow = prefect.Flow(**flow_kwargs)
        tasks: dict[Task, prefect.Task] = {}

        for t in f.get_tasks():
            task = prefect.task(name=t.name)(t.run)
            tasks[t] = task

            # if desired input is passed in inputs use it,
            # otherwise try to get it from results
            task_inputs = {
                **{in_id: tasks[in_t] for in_id, in_t in t.input_tasks.items() if in_t in tasks and in_t not in inputs},
                **{in_id: Table(inputs[in_t]) for in_id, in_t in t.input_tasks.items() if in_t in inputs},
            }

            flow.add_task(task)
            flow.set_dependencies(
                task,
                keyword_tasks=dict(
                    inputs=task_inputs,
                    run_context=run_context,
                    config_context=config_context,
                ),
            )

        for task in f.get_tasks():
            for parent in f.get_parent_tasks(task):
                flow.add_edge(tasks[parent], tasks[task])

        project_name = config_context.pipedag_name + "-" + config_context.instance_id
        try:
            flow.register(project_name=project_name)
        except ValueError as _e:
            self.logger.warning(f"Please make sure project {project_name} exists: {_e}")

        return flow, tasks

    def run(
        self,
        flow: Subflow,
        ignore_position_hashes: bool = False,
        inputs: dict[Task, ExternalTableReference] | None = None,
        **run_kwargs,
    ):
        _ = ignore_position_hashes
        prefect_flow, tasks_map = self.construct_prefect_flow(flow, inputs)
        result = prefect_flow.run(**run_kwargs)

        # Compute task_values
        task_values = {}
        for task, prefect_task in tasks_map.items():
            task_values[task] = result.result[prefect_task].result

        # If the task failed, extract the exception
        exception = None
        if result.is_failed():
            for task_res in result.result.values():
                if task_res.is_failed() and isinstance(task_res.result, Exception):
                    exception = task_res.result
                    break
            else:
                # Generic Fallback
                exception = Exception(f"Prefect run failed with message: {result.message}")

        return Result.init_from(
            subflow=flow,
            underlying=result,
            successful=result.is_successful(),
            task_values=task_values,
            exception=exception,
        )


@requires(prefect, ImportError("Module 'prefect' not installed"))
@requires(
    prefect_version in SpecifierSet(">=2.0,<4.0"),
    ImportWarning(f"Requires prefect version 2.x or 3.x (found {prefect_version})"),
)
class PrefectTwoEngine(OrchestrationEngine):
    """
    Hands over execution of a flow to `Prefect 2 <https://docs.prefect.io>`_.

    :param flow_kwargs:
        Optional dictionary of keyword arguments that get passed to the
        initializer of |@prefect2.flow|_ decorator.

    .. |@prefect2.flow| replace:: ``@prefect.flow``
    .. _@prefect2.flow:
            https://docs.prefect.io/latest/api-ref/prefect/flows/#prefect.flows.flow
    """

    def __init__(self, flow_kwargs: dict[str, Any] = None):
        self.flow_kwargs = flow_kwargs or {}

    def construct_prefect_flow(
        self,
        f: Subflow,
        inputs: dict[Task, ExternalTableReference] | None = None,
    ) -> prefect.Flow:
        from pydiverse.pipedag.materialize.materializing_task import MaterializingTask

        inputs = inputs if inputs is not None else {}

        run_context = RunContext.get()
        config_context = ConfigContext.get()

        flow_kwargs = {
            "name": f.name,
            "validate_parameters": False,
            **self.flow_kwargs,
        }

        @prefect.flow(**flow_kwargs)
        def pipedag_flow():
            futures: dict[Task, prefect.futures.PrefectFuture] = {}

            for t in f.get_tasks():
                task_kwargs = {"name": t.name}
                if isinstance(t, MaterializingTask):
                    task_kwargs["version"] = t.version

                task = prefect.task(**task_kwargs)(t.run)

                parents = [futures[p] for p in f.get_parent_tasks(t)]
                task_inputs = {
                    **{
                        in_id: futures[in_t]
                        for in_id, in_t in t.input_tasks.items()
                        if in_t in futures and in_t not in inputs
                    },
                    **{in_id: Table(inputs[in_t]) for in_id, in_t in t.input_tasks.items() if in_t in inputs},
                }

                futures[t] = task.submit(
                    inputs=task_inputs,
                    run_context=run_context,
                    config_context=config_context,
                    wait_for=parents,
                )

            return futures

        return pipedag_flow

    def run(
        self,
        flow: Subflow,
        ignore_position_hashes: bool = False,
        inputs: dict[Task, ExternalTableReference] | None = None,
        **kwargs,
    ):
        _ = ignore_position_hashes
        if kwargs:
            raise TypeError(f"{type(self).__name__}.run doesn't take kwargs.")
        prefect_flow = self.construct_prefect_flow(flow, inputs)
        result = prefect_flow(return_state=True)

        # Compute task_values
        task_values = {}
        successful = result.is_completed()

        for task, state in result.result().items():
            if state.is_completed():
                task_values[task] = state.result()
            else:
                successful = False

        # If the task failed, extract the exception
        exception = None
        if not successful:
            for state in result.result().values():
                if state.is_failed() or state.is_crashed():
                    exception = prefect.states.get_state_exception(state)
                    break

        return Result.init_from(
            subflow=flow,
            underlying=result,
            successful=successful,
            task_values=task_values,
            exception=exception,
        )


# Automatic Prefect Version Selection


if prefect_version in SpecifierSet("~=1.0"):
    PrefectEngine = PrefectOneEngine
elif prefect_version in SpecifierSet(">=2.0,<4.0"):
    # prefect 2 and 3 are not that different
    PrefectEngine = PrefectTwoEngine
else:
    from abc import ABC

    @requires(prefect, ImportWarning("Module 'prefect' not installed"))
    @requires(False, ImportWarning(f"Incompatible prefect version {prefect_version}"))
    class PrefectEngine(OrchestrationEngine, ABC):
        pass



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from io import BytesIO

# Patch pytest EncodedFile (from pytest-capture plugin) to be pickleable
# https://github.com/mariusvniekerk/pytest-dask/blob/master/pytest_dask/serde_patch.py
from _pytest.capture import EncodedFile


def apply_getsetstate(cls):
    def inner(ref):
        cls.__getstate__ = ref.__getstate__
        cls.__reduce__ = ref.__reduce__
        cls.__reduce_ex__ = ref.__reduce_ex__
        return cls

    return inner


@apply_getsetstate(EncodedFile)
class _EncodedFile:
    def __getstate__(self):
        assert isinstance(self, EncodedFile)
        current_position = self.buffer.seek(0, 1)
        self.buffer.seek(0)
        value = self.buffer.read()
        self.buffer.seek(current_position, 0)
        return {"value": value, "encoding": self.encoding}

    def __reduce__(self):
        state = self.__getstate__()
        return self.__class__, (BytesIO(state["value"]), state["encoding"])

    def __reduce_ex__(self, protocol):
        _ = protocol
        return self.__reduce__()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import polars as pl
import sqlalchemy as sa

from pydiverse.pipedag import ConfigContext, Flow, Stage, Table, materialize
from tests.fixtures.instances import with_instances


@with_instances(
    "local_table_cache",
    "local_table_cache_inout",
    "local_table_cache_inout_numpy",
    "local_table_store",
)
def test_local_table_cache(mocker):
    input_val_ = 0

    @materialize()
    def input_val():
        return input_val_

    @materialize(version="1.0")
    def select_pandas(x):
        # Supported by local caching
        return Table(pd.DataFrame({"x": [x]}), "pandas")

    @materialize(version="1.0")
    def select_polars(x):
        # Supported by local caching
        return Table(pl.DataFrame({"x": [x]}), "polars")

    @materialize(lazy=True)
    def select_sql(x):
        # Not supported by local caching
        return Table(sa.select(sa.literal(x).label("x")), "sql")

    @materialize(version="1.0", input_type=pd.DataFrame)
    def sink_pandas(*args):
        for arg in args:
            assert arg["x"][0] == input_val_

    @materialize(version="1.0", input_type=pl.DataFrame)
    def sink_polars(*args):
        for arg in args:
            assert arg["x"][0] == input_val_

    @materialize(version="1.0", input_type=sa.Table)
    def sink_sql(*args):
        for arg in args:
            # check that column x exists
            _ = arg.c.x

    with Flow() as f:
        with Stage("stage"):
            x = input_val()

            s_pandas = select_pandas(x)
            s_polars = select_polars(x)
            s_sql = select_sql(x)

            _ = sink_pandas(s_pandas, s_polars, s_sql)
            _ = sink_polars(s_pandas, s_polars, s_sql)
            _ = sink_sql(s_pandas, s_polars, s_sql)

    # Initial run to invalidate cache
    input_val_ = -1
    f.run()
    input_val_ = 0

    # Spy Setup
    config_context = ConfigContext.get()
    local_table_cache = config_context.store.local_table_cache

    si = int(local_table_cache.should_store_input)
    so = int(local_table_cache.should_store_output)
    siac = int(local_table_cache.should_use_stored_input_as_cache)

    store_table_spy = mocker.spy(local_table_cache, "store_table")
    store_input_spy = mocker.spy(local_table_cache, "store_input")
    _store_table_spy = mocker.spy(local_table_cache, "_store_table")

    retrieve_table_obj_spy = mocker.spy(local_table_cache, "retrieve_table_obj")
    _retrieve_table_obj_spy = mocker.spy(local_table_cache, "_retrieve_table_obj")

    # Initial Run
    f.run()

    expected_retrieve_table_obj = 3 * 3
    expected_successful_retrieve_table_obj = 2 * so * siac + 3 * siac
    expected_store_table = 3
    expected_store_input = expected_store_table * 3 - expected_successful_retrieve_table_obj

    assert store_table_spy.call_count == expected_store_table
    assert store_input_spy.call_count == expected_store_input
    assert retrieve_table_obj_spy.call_count == expected_retrieve_table_obj

    assert _store_table_spy.call_count == (expected_store_input * si) + (expected_store_table * so)
    assert _retrieve_table_obj_spy.call_count == expected_successful_retrieve_table_obj + siac * expected_store_table

    # Second Run
    store_table_spy.reset_mock()
    store_input_spy.reset_mock()
    _store_table_spy.reset_mock()
    retrieve_table_obj_spy.reset_mock()
    _retrieve_table_obj_spy.reset_mock()

    f.run()

    # Everything should be cache valid, thus no task should get executed.
    assert store_table_spy.call_count == 0
    assert store_input_spy.call_count == 0
    assert retrieve_table_obj_spy.call_count == 0



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import sqlalchemy as sa
from pandas.testing import assert_frame_equal

from pydiverse.pipedag import Blob, Flow, Stage, Table, materialize
from pydiverse.pipedag.context import StageLockContext
from tests.fixtures.instances import with_instances

dfA = pd.DataFrame(
    {
        "a": [0, 1, 2, 4],
        "b": [9, 8, 7, 6],
    }
)

dfB = pd.DataFrame(
    {
        "a": [2, 1, 0, 1],
        "x": [1, 1, 2, 2],
    }
)


@materialize(nout=2, version="1")
def inputs():
    import time

    time.sleep(1)
    return Table(dfA, "dfA"), Table(dfB, "dfB_%%")


@materialize(input_type=pd.DataFrame)
def double_values(df: pd.DataFrame):
    return Table(df.transform(lambda x: x * 2))


@materialize(input_type=sa.Table, lazy=True)
def join_on_a(left: sa.sql.expression.Alias, right: sa.sql.expression.Alias):
    return Table(left.select().join(right, left.c.a == right.c.a))


@materialize(input_type=pd.DataFrame)
def list_arg(x: list[pd.DataFrame]):
    assert isinstance(x[0], pd.DataFrame)
    return Blob(x)


@materialize
def blob_task(x, y):
    return Blob(x), Blob(y)


@with_instances("postgres")
def test_simple_flow(with_blob=True):
    with Flow() as flow:
        with Stage("simple_flow_stage1"):
            inp = inputs()
            a, b = inp

            a2 = double_values(a)
            b2 = double_values(b)
            b4 = double_values(b2)
            b4 = double_values(b4)
            x = list_arg([a2, b, b4])

        with Stage("simple_flow_stage2"):
            joined = join_on_a(a2, b4)
            joined_times_2 = double_values(joined)

            if with_blob:
                v = blob_task(x, x)
                v = blob_task(v, v)
                v = blob_task(v, v)

                blob_tuple = blob_task(1, 2)

    with StageLockContext():
        result = flow.run()  # this will use the default configuration instance=__any__
        assert result.successful

        # Check result.get works
        res_a = result.get(a, as_type=pd.DataFrame)
        res_b = result.get(b, as_type=pd.DataFrame)
        res_inp = result.get(inp, as_type=pd.DataFrame)
        res_joined = result.get(joined, as_type=pd.DataFrame)
        res_joined_times_2 = result.get(joined_times_2, as_type=pd.DataFrame)

        assert_frame_equal(res_a, dfA, check_dtype=False)
        assert_frame_equal(res_b, dfB, check_dtype=False)
        assert_frame_equal(res_inp[0], dfA, check_dtype=False)
        assert_frame_equal(res_inp[1], dfB, check_dtype=False)
        assert_frame_equal(res_joined * 2, res_joined_times_2)

        result.get(x)
        if with_blob:
            result.get(v)
            assert tuple(result.get(blob_tuple)) == (1, 2)


if __name__ == "__main__":
    test_simple_flow()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import traceback

import pytest

from pydiverse.common.util import Disposable, requires
from pydiverse.pipedag.errors import DisposedError


def test_requires():
    @requires(None, ImportError("Some Error"))
    class BadClass:
        a = 1
        b = 2

    # Shouldn't be able to create instance
    with pytest.raises(ImportError, match="Some Error"):
        BadClass()

    # Shouldn't be able to access class attribute
    with pytest.raises(ImportError, match="Some Error"):
        _ = BadClass.a

    # If all requirements are fulfilled, nothing should change
    @requires((pytest,), Exception("This shouldn't happen"))
    class GoodClass:
        a = 1

    _ = GoodClass()
    _ = GoodClass.a


def test_disposable():
    class Foo(Disposable):
        a = 1

        def bar(self):
            return 2

    x = Foo()

    assert x.a == 1
    assert x.bar() == 2

    x.dispose()

    with pytest.raises(DisposedError):
        _ = x.a
    with pytest.raises(DisposedError):
        x.foo()
    with pytest.raises(DisposedError):
        x.dispose()
    with pytest.raises(DisposedError):
        x.a = 1


def test_format_exception():
    # traceback.format_exception syntax changed from python 3.9 to 3.10
    # thus we use traceback.format_exc()
    try:
        raise RuntimeError("this error is intended by test")
    except RuntimeError:
        trace = traceback.format_exc()
        assert 'RuntimeError("this error is intended by test")' in trace
        assert "test_util.py" in trace



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import sys
from abc import ABC, abstractmethod
from collections.abc import Iterable
from enum import Enum
from typing import Any, Generic

from typing_extensions import Self

from pydiverse.common.util import requires
from pydiverse.common.util.computation_tracing import ComputationTracer
from pydiverse.pipedag import Table
from pydiverse.pipedag._typing import T, TableHookResolverT
from pydiverse.pipedag.context import RunContext
from pydiverse.pipedag.errors import StoreIncompatibleException
from pydiverse.pipedag.materialize.materializing_task import (
    AutoVersionSupport,
    MaterializingTask,
)


class CanMatResult(Enum):
    NO = 0
    YES = 1
    # Don't assume that other tables of same obj type can be materialized the same way
    YES_BUT_DONT_CACHE = 2

    @staticmethod
    def new(bool_val: bool):
        return CanMatResult.YES if bool_val else CanMatResult.NO


class CanRetResult(Enum):
    NO = 0
    YES = 1
    # For store and type combination it is expected that no hook can retrieve
    NO_HOOK_IS_EXPECTED = 2

    @staticmethod
    def new(bool_val: bool):
        return CanRetResult.YES if bool_val else CanRetResult.NO


class TableHook(Generic[TableHookResolverT], ABC):
    """Base class to define how to handle a specific table

    For more information, take a look at the `BaseTableStore.register_table`
    documentation.
    """

    __slots__ = ()
    auto_version_support: AutoVersionSupport = AutoVersionSupport.NONE

    @classmethod
    @abstractmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        """
        Return `True` if this hook can materialize a table with the specified
        underlying type. If `True` is returned, the `materialize` method
        MUST be implemented for the type.
        """

    @classmethod
    @abstractmethod
    def can_retrieve(cls, type_: type) -> CanRetResult:
        """
        Return `CanRetResult.YES` if this hook can retrieve a table from the store
        and convert it to the specified type. If `CanRetResult.YES` is returned, the
        `retrieve` method MUST be implemented for the type.

        If `CanRetResult.NO_HOOK_IS_EXPECTED` is returned, it is expected that no
        hook can retrieve the table as the type. This is used in parquet table cache
        which does not make sense for SQL references to suppress a warning.
        """

    @classmethod
    def retrieve_as_reference(cls, type_: type) -> bool:
        """
        Table hooks that retrieve `type_` as reference instead of copying the content
        of the table should return True. This is used by imperative materialization
        to determine whether input_type is a good default for returned references.
        """
        return False

    @classmethod
    @abstractmethod
    def materialize(
        cls,
        store: TableHookResolverT,
        table: Table,
        stage_name: str,
        without_config_context: bool = False,
    ) -> None:
        """Materialize a table object

        :param store: The store which called this method
        :param table: The table that should be materialized
        :param stage_name: The name of the stage in which the table should
            be stored - can either be `stage.name` or `stage.transaction_name`.
        """

    @classmethod
    @abstractmethod
    def retrieve(
        cls,
        store: TableHookResolverT,
        table: Table,
        stage_name: str | None,
        as_type: type[T] | tuple | dict[str, Any],
        limit: int | None = None,
    ) -> T:
        """Retrieve a table from the store

        :param store: The store in which the table is stored
        :param table: The table which should get retrieved
        :param stage_name: The name of the stage from which te table should
            be retrieved
        :param as_type: The type as which the table is to be retrieved
        :param limit: The maximum number of rows to retrieve
            (won't work for SQL references types)
        :return: The retrieved table (converted to the correct type)
        """

    @classmethod
    def auto_table(cls, obj: T) -> Table[T]:
        """Wrap an object inside a `Table`

        Given an object that can be materialized by this hook, produces a
        Table instance that may contain additional metadata. This is useful
        to enable automatic name propagation.

        :param obj: The object which should get wrapped inside a `Table`
        :return: The `Table` object which wraps `obj`
        """
        return Table(obj)

    @classmethod
    def lazy_query_str(cls, store: TableHookResolverT, obj) -> str:
        """String that represents the associated object

        Can either be a literal query string (e.g. SQL query), or a string that
        uniquely represents the query in some other way. Used for computing
        the cache key for materializing tasks with `lazy=True`.

        :raises TypeError: if the type doesn't support lazy queries.
        """
        raise TypeError(f"Lazy query not supported with object of type {type(obj)}")

    @classmethod
    def retrieve_for_auto_versioning_lazy(
        cls,
        store: TableHookResolverT,
        table: Table,
        stage_name: str,
        as_type: type[T] | tuple | dict[str, Any],
    ) -> T:
        """
        Retrieve a table (or table like object) from the store to be used for
        automatic version number determination.

        This function gets called with the same arguments as ``.retrieve``.

        Must be implemented for ``auto_version_support == AutoVersionSupport.LAZY``.

        :raises TypeError: if the type doesn't support automatic versioning.
        """
        raise TypeError(f"Auto versioning not supported for type {as_type}")

    @classmethod
    def get_auto_version_lazy(cls, obj) -> str:
        """
        Must be implemented for ``auto_version_support == AutoVersionSupport.LAZY``.

        :param obj: object returned from task
        :return: string representation of the operations performed on this object.
        :raises TypeError: if the object doesn't support automatic versioning.
        """
        raise NotImplementedError

    @classmethod
    def get_computation_tracer(cls) -> ComputationTracer:
        """
        Get a computation tracer for this type of task.
        Must be implemented for ``auto_version_support == AutoVersionSupport.TRACE``.
        """
        raise NotImplementedError


class TableHookResolver:
    class ClassState:
        def __init__(self):
            self.registered_table_hooks: list[type[TableHook]] = []
            self.m_hook_cache: dict[type, type[TableHook]] = {}
            self.r_hook_cache: dict[type, type[TableHook]] = {}
            self.hook_subclass_cache: dict[type, type[TableHook]] = {}

        def add_hook(self, hook_cls, replace_hooks: list[type[TableHook]] | None):
            # we ignore if replace_hooks are not found since they might be found in
            # base class with lower lookup priority
            if replace_hooks:
                self.registered_table_hooks = [x for x in self.registered_table_hooks if x not in replace_hooks]
            self.registered_table_hooks.append(hook_cls)

        def __repr__(self):
            return "ClassState=" + str(
                {x: getattr(self, x) for x in dir(self) if not x.startswith("_") and x != "add_hook"}
            )

    _states: dict[int, ClassState] = {}

    @classmethod
    def _resolver_state(cls: "TableHookResolver"):
        state = cls._states.get(id(cls))
        if state is None:
            state = cls._states[id(cls)] = cls.ClassState()
        return state

    @classmethod
    def register_table(
        cls: "TableHookResolver",
        *requirements: Any,
        replace_hooks: list[type[TableHook]] | None = None,
    ):
        """Decorator to register a `TableHook`

        Each table store should be able to handle tables of various different
        types (e.g. a pandas dataframe, a sqlalchemy select statement, a
        pydiverse transform table, ...). To add table type specific logic to
        a table store so called TableHooks are used.

        This decorator is used to register such a hook::

            try:
                from library import x
            except ImportError as e:
                warnings.warn(str(e), ImportWarning)
                x = None

            @TableStore.register_table(x)
            def XTableHook(TableHook):
                ...

        :param requirements: The requirements which must be satisfied to register
            the decorated class.
        :param replace_hooks: Takes a list of TableHook classes (e.g. PandasTableHook)
            that should be replaced by the decorated class.
            If None, the new hook is just added to the list of available hooks.

            Details: specifying replace_hooks helps also implicitly because the
            mentioned hooks must have been imported/registered before. If they were
            registered on a base class of the hook resolver (i.e. table store) of this
            call, they will be ignored but hooks registered on subclasses take
            precedence anyway. Please make sure not provide replace hooks registered to
            subclasses of this call's hook resolver. This will fail.
        """

        def decorator(hook_cls):
            if not all(requirements):
                cls.__hooks_with_unmet_requirements.append(hook_cls.__module__ + "." + hook_cls.__qualname__)
                return requires(
                    False,
                    Exception(f"Not all requirements met for {hook_cls.__name__}"),
                )(hook_cls)

            cls._resolver_state().add_hook(hook_cls, replace_hooks)
            return hook_cls

        return decorator

    def __all_registered_table_hooks(self) -> Iterable[type[TableHook]]:
        # Walk up the hierarchy of super classes and return each registered
        # table hook in reverse order
        for cls in type(self).__mro__:
            if issubclass(cls, TableHookResolver):
                yield from cls._resolver_state().registered_table_hooks[::-1]  # noqa

    def get_m_table_hook(self: Self, tbl: Table) -> type[TableHook[Self]]:
        """Get a table hook that can materialize the specified type"""
        type_ = type(tbl.obj)
        if type_ in self._resolver_state().m_hook_cache:
            return self._resolver_state().m_hook_cache[type_]

        for hook in self.__all_registered_table_hooks():
            can = hook.can_materialize(tbl)
            if can != CanMatResult.NO and can is not False:
                if can == CanMatResult.YES:
                    self._resolver_state().m_hook_cache[type_] = hook
                return hook

        raise TypeError(
            f"Can't materialize Table with underlying type {type_}. " + self.__hook_unmet_requirements_message()
        )

    def get_r_table_hook(self: Self, type_: type[T] | tuple | dict) -> type[TableHook[Self]] | None:
        """Get a table hook that can retrieve the specified type"""
        if isinstance(type_, tuple):
            type_ = type_[0]
        elif isinstance(type_, dict):
            type_ = type_["type"]

        if type_ is None:
            raise ValueError("type_ argument can't be None")

        if type_ in self._resolver_state().r_hook_cache:
            return self._resolver_state().r_hook_cache[type_]

        for hook in self.__all_registered_table_hooks():
            can_ret = hook.can_retrieve(type_)
            if can_ret == CanRetResult.YES:
                self._resolver_state().r_hook_cache[type_] = hook
                return hook
            elif can_ret == CanRetResult.NO_HOOK_IS_EXPECTED:
                raise StoreIncompatibleException(
                    f"Can't retrieve Table as type {type_}. This type is incompatible with store {self}."
                )

        raise TypeError(f"Can't retrieve Table as type {type_}. " + self.__hook_unmet_requirements_message())

    def get_hook_subclass(self: Self, type_: type[T]) -> type[T]:
        """Finds a table hook that is a subclass of the provided type"""
        if type_ in self._resolver_state().hook_subclass_cache:
            return self._resolver_state().hook_subclass_cache[type_]

        min_distance = sys.maxsize
        min_subclass = None

        for cls in type(self).__mro__:
            if issubclass(cls, TableHookResolver):
                for hook in cls._resolver_state().registered_table_hooks[::-1]:  # noqa
                    if issubclass(hook, type_):
                        # Calculate distance between superclass and child class
                        try:
                            dist = hook.__mro__.index(type_)
                            if dist < min_distance:
                                min_distance = dist
                                min_subclass = hook
                        except ValueError:
                            pass

                if min_subclass:
                    break

        if min_subclass:
            self._resolver_state().hook_subclass_cache[type_] = min_subclass
            return min_subclass

        raise RuntimeError

    def store_table(self, table: Table, task: MaterializingTask | None):
        """Stores a table in the associated transaction stage

        The store must convert the table object (`table.obj`) to the correct
        internal type. This means, that in some cases it first has to
        evaluate a lazy object. For example: if a sql based table store
        receives a sql query to store, it has to execute it first.

        The implementation details of this get handled by the registered
        TableHooks.
        """

        # In case of deferred operations, inform run context that stage
        # isn't 100% cache valid anymore.
        if task is not None:
            RunContext.get().set_stage_has_changed(task.stage)

        # Materialize
        hook = self.get_m_table_hook(table)
        hook.materialize(self, table, table.stage.transaction_name)

    def retrieve_table_obj(
        self,
        table: Table,
        as_type: type[T],
        for_auto_versioning: bool = False,
    ) -> T:
        """Loads a table from the store

        Retrieves the table from the store, converts it to the correct
        If the stage hasn't yet been committed, the table must be retrieved
        from the transaction, else it must be retrieved from the committed
        stage.

        :raises TypeError: if the retrieved table can't be converted to
            the requested type.
        """

        if as_type is None:
            raise TypeError("Missing 'as_type' argument. You must specify a type to be able to dematerialize a Table.")

        hook = self.get_r_table_hook(as_type)
        stage_name = table.stage.current_name if table.stage is not None else None
        try:
            if for_auto_versioning:
                return hook.retrieve_for_auto_versioning_lazy(self, table, stage_name, as_type)

            return hook.retrieve(self, table, stage_name, as_type)
        except Exception as e:
            raise RuntimeError(f"Failed to retrieve table '{table}'") from e

    __hooks_with_unmet_requirements: list[str] = []

    @classmethod
    def __hook_unmet_requirements_message(cls) -> str:
        if len(cls.__hooks_with_unmet_requirements) == 0:
            return "This is is because no TableHook has been registered for this type."

        return (
            "This is either because no TableHook has been registered for this type, "
            "or because not all requirements have been met for the corresponding hook."
            "\nHooks with unmet requirements: " + ", ".join(cls.__hooks_with_unmet_requirements)
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import threading
import time
from typing import Callable

import pytest
import structlog

from pydiverse.pipedag.backend.lock import BaseLockManager, LockState
from pydiverse.pipedag.backend.lock.zookeeper import KazooClient
from pydiverse.pipedag.util.timing import timeout
from tests.fixtures.instances import with_instances


class RaisingThread(threading.Thread):
    def run(self) -> None:
        try:
            super().run()
        except BaseException as e:
            self.exception = e

    def join(self, timeout=None) -> None:
        super().join(timeout)
        if e := getattr(self, "exception", None):
            raise e


def _test_lock_manager(create_lock_manager: Callable[[], BaseLockManager]):
    lock_name = "test_lock"
    sleep_time = 1

    ready_barrier = threading.Barrier(2)
    locked_event = threading.Event()
    sleep_event = threading.Event()
    unlocked_event = threading.Event()
    done_barrier = threading.Barrier(2)

    def lm_1_task():
        lm = create_lock_manager()
        ready_barrier.wait()

        try:
            with lm(lock_name):
                assert lm.get_lock_state(lock_name) == LockState.LOCKED
                locked_event.set()
                time.sleep(sleep_time)
                sleep_event.set()
                time.sleep(0.025)
            unlocked_event.set()
        finally:
            done_barrier.wait(timeout=3)
            assert lm.get_lock_state(lock_name) == LockState.UNLOCKED
            lm.dispose()

    def lm_2_task():
        lm = create_lock_manager()
        ready_barrier.wait()

        locked_event.wait()
        start_time = time.perf_counter()

        try:
            with lm(lock_name):
                assert lm.get_lock_state(lock_name) == LockState.LOCKED
                end_time = time.perf_counter()

                if not sleep_event.is_set() or not unlocked_event.wait(timeout=0.025):
                    raise RuntimeError(
                        "Second lock manager was able to acquire lock before the first lock manager released it."
                    )

            delta = end_time - start_time
            assert delta >= sleep_time
        finally:
            done_barrier.wait(timeout=3)
            assert lm.get_lock_state(lock_name) == LockState.UNLOCKED
            lm.dispose()

    t1 = RaisingThread(target=lm_1_task, daemon=True)
    t2 = RaisingThread(target=lm_2_task, daemon=True)

    t1.start()
    t2.start()

    t1.join(timeout=(5 + sleep_time))
    assert not t1.is_alive(), "Thread timed out"

    t2.join(timeout=5)
    assert not t2.is_alive(), "Thread timed out"


@pytest.mark.parallelize
@pytest.mark.skipif(KazooClient is None, reason="requires kazoo")
def test_zookeeper():
    from pydiverse.pipedag.backend.lock import ZooKeeperLockManager

    def create_lock_manager():
        # TODO: Don't hardcode ip - Get somehow from the environment
        client = KazooClient(hosts="localhost:2181")
        return ZooKeeperLockManager(client, "pipedag/tests/zookeeper/")

    try:
        with timeout(seconds=60):
            _test_lock_manager(create_lock_manager)
    except TimeoutError:
        logger = structlog.get_logger(logger_name=__name__ + ".test_zookeeper")
        logger.info(
            "TODO: It is not understood why test_zookeeper can deadlock; we are already using `Thread.join(timeout)"
        )


@pytest.mark.parallelize
def test_filelock():
    import tempfile
    from pathlib import Path

    from pydiverse.pipedag.backend.lock import FileLockManager

    base_path = Path(tempfile.gettempdir()) / "pipedag" / "tests"

    def create_lock_manager():
        return FileLockManager(base_path=base_path)

    _test_lock_manager(create_lock_manager)


@pytest.mark.parallelize
def test_no_lock():
    from pydiverse.pipedag.backend.lock import NoLockManager

    def create_lock_manager():
        return NoLockManager()

    try:
        with timeout(seconds=60):
            with pytest.raises(RuntimeError):
                _test_lock_manager(create_lock_manager)
                pytest.fail("No lock manager MUST fail the lock manager tests")
    except TimeoutError:
        logger = structlog.get_logger(logger_name=__name__ + ".test_no_lock")
        logger.info("Running without lock manage may hang")


@with_instances("postgres", "mssql", "ibm_db2")
def test_database():
    from pydiverse.pipedag import ConfigContext
    from pydiverse.pipedag.backend.lock import DatabaseLockManager

    config_context = ConfigContext.get()

    def create_lock_manager():
        with config_context.evolve():
            return DatabaseLockManager.init_from_config_context(config_context)

    _test_lock_manager(create_lock_manager)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy

import pandas as pd
import pytest
import sqlalchemy as sa

from pydiverse.pipedag import ConfigContext, Flow, Stage, StageLockContext, Table
from pydiverse.pipedag.container import Schema
from pydiverse.pipedag.core.config import PipedagConfig, create_basic_pipedag_config
from tests.fixtures.instances import with_instances
from tests.util import tasks_library as m
from tests.util import tasks_library_imperative as m2


@with_instances(
    "postgres",
    "local_table_cache",
    "local_table_cache_inout",
    "local_table_cache_inout_numpy",
)
@pytest.mark.parametrize("imperative", [False, True])
def test_get_output_from_store(imperative):
    _m = m2 if imperative else m
    with Flow() as f:
        with Stage("stage_1"):
            df1 = _m.pd_dataframe({"x": [0, 1, 2, 3]})
            df2 = _m.pd_dataframe({"y": [0, 1, 2, 3]})
            dataframes = _m.create_tuple(df1, df2)

    # We only use the StageLockContext for testing
    with StageLockContext():
        result = f.run()

        # Call on MaterializingTask
        pd.testing.assert_frame_equal(
            df1.get_output_from_store(as_type=pd.DataFrame),
            result.get(df1, as_type=pd.DataFrame),
        )

        pd.testing.assert_frame_equal(
            df2.get_output_from_store(as_type=pd.DataFrame),
            result.get(df2, as_type=pd.DataFrame),
        )

        pd.testing.assert_frame_equal(
            dataframes.get_output_from_store(as_type=pd.DataFrame)[1],
            result.get(dataframes, as_type=pd.DataFrame)[1],
        )

        # since df1 and df2 are outputs from tasks with the same name (pd_dataframe)
        # ignoring the position hash when retrieving
        # should just return the output from the latest call
        pd.testing.assert_frame_equal(
            df1.get_output_from_store(as_type=pd.DataFrame, ignore_position_hashes=True),
            result.get(df2, as_type=pd.DataFrame),
        )

        # Call on MaterializingTaskGetItem
        pd.testing.assert_frame_equal(
            dataframes[0].get_output_from_store(as_type=pd.DataFrame),
            dataframes.get_output_from_store(as_type=pd.DataFrame)[0],
        )

        pd.testing.assert_frame_equal(
            dataframes[0].get_output_from_store(as_type=pd.DataFrame, ignore_position_hashes=True),
            dataframes.get_output_from_store(as_type=pd.DataFrame, ignore_position_hashes=True)[0],
        )


@with_instances("postgres")
@pytest.mark.parametrize("imperative", [False, True])
def test_call_task_outside_flow(imperative):
    _m = m2 if imperative else m
    # Literal Values
    assert m.one() == 1
    assert m.create_tuple(m.one(), m.two()) == (1, 2)
    assert m.noop("foo") == "foo"

    # Table
    df1 = _m.pd_dataframe({"x": [0, 1, 2, 3]})
    df2 = m.noop(df1)
    expected = pd.DataFrame({"x": [0, 1, 2, 3]})

    for df in [df1, df2]:
        pd.testing.assert_frame_equal(df, expected)
        _m.assert_table_equal(df, expected)

    # Lazy Table
    expr = _m.simple_lazy_table()
    expected_expr = (
        "SELECT 0 AS col1, '0' AS col2 "
        "UNION ALL SELECT 1 AS col1, '1' AS col2 "
        "UNION ALL SELECT 2 AS col1, '2' AS col2 "
        "UNION ALL SELECT 3 AS col1, '3' AS col2"
    )
    config_context = ConfigContext.get()
    assert (
        str(
            expr.compile(
                config_context.store.table_store.engine,
                compile_kwargs={"literal_binds": True},
            )
        )
        == expected_expr
    )

    tbl = m.noop(sa.Table("foo", sa.MetaData()).alias("bar"))
    assert tbl.name == "bar"
    assert tbl.original.name == "foo"

    # Blob
    expected_blob = {"a": {"b": 1, "c": 2}, "d": [1, (2, 3)]}
    blob = m.as_blob(expected_blob)

    assert blob is expected_blob
    m.assert_blob_equal(blob, expected_blob)

    # RawSql
    expected_query = "SELECT 1 as x"
    raw_sql = m.as_raw_sql(expected_query)

    assert raw_sql == expected_query


@with_instances("postgres")
def test_imperative_materialize_given_config():
    _m = m2
    df = pd.DataFrame({"x": [0, 1, 2, 3]})
    tbl = Table(df)
    assert tbl.materialize() is df  # should be no-op
    assert tbl.materialize(return_nothing=True) is None

    config_context = ConfigContext.get()
    engine = config_context.store.table_store.engine
    with engine.connect() as conn:
        conn.execute(sa.text("DROP SCHEMA IF EXISTS dummy_schema CASCADE"))
        conn.execute(sa.text("CREATE SCHEMA IF NOT EXISTS dummy_schema"))
        if sa.__version__ >= "2.0.0":
            conn.commit()

    # test failure on materialization without schema
    with pytest.raises(ValueError, match="schema must be provided"):
        tbl.materialize(config_context)
    # test materialization with explicit config_context and direct schema
    ref = tbl.materialize(config_context, schema=Schema("dummy_schema"))
    sa_tbl = sa.Table(ref.original.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    sa_tbl2 = _m.noop_lazy(sa_tbl.alias("alias"))
    assert sa_tbl2.name == "alias"
    assert sa_tbl2.original.name == sa_tbl.name

    # test return nothing materialization
    tbl.name = "a"
    assert tbl.materialize(config_context, schema=Schema("dummy_schema"), return_nothing=True) is None
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context and prefixed schema
    tbl.name = "b"
    tbl.materialize(config_context, schema=Schema("schema", prefix="dummy_"))
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context and postfixed schema
    tbl.name = "c"
    tbl.materialize(config_context, schema=Schema("dummy", suffix="_schema"))
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context incl. prefixed schema
    pipedag_config = create_basic_pipedag_config(
        config_context.store.table_store.engine.url.render_as_string(hide_password=False)
    )
    raw_config2 = copy.deepcopy(pipedag_config.raw_config)
    raw_config2["instances"]["__any__"]["table_store"]["args"]["schema_prefix"] = "dummy_"
    cfg2 = PipedagConfig(raw_config2).get()
    tbl.name = "e"
    with pytest.raises(ValueError, match="Schema prefix and postfix must match"):
        tbl.materialize(cfg2, schema=Schema("schema", prefix="dummyX_"))
    tbl.materialize(cfg2, schema=Schema("schema", prefix="dummy_"))
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context and postfixed schema
    pipedag_config = create_basic_pipedag_config(
        config_context.store.table_store.engine.url.render_as_string(hide_password=False)
    )
    raw_config2 = copy.deepcopy(pipedag_config.raw_config)
    raw_config2["instances"]["__any__"]["table_store"]["args"]["schema_suffix"] = "_schema"
    cfg2 = PipedagConfig(raw_config2).get()
    tbl.name = "e"
    with pytest.raises(ValueError, match="Schema prefix and postfix must match"):
        tbl.materialize(cfg2, schema=Schema("dummy", suffix="_Xschema"))
    tbl.materialize(cfg2, schema=Schema("dummy", suffix="_schema"))
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import json
import os
import shutil
import types
from typing import Any

import pandas as pd
import sqlalchemy as sa
from packaging.version import Version
from upath import UPath

import pydiverse.pipedag.backend.table.sql.hooks as sql_hooks
from pydiverse.pipedag import ConfigContext, Stage, Table
from pydiverse.pipedag.materialize.materializing_task import MaterializingTask
from pydiverse.pipedag.materialize.store import BaseTableCache
from pydiverse.pipedag.materialize.table_hook_base import CanMatResult, CanRetResult, TableHook
from pydiverse.pipedag.util import normalize_name
from pydiverse.pipedag.util.path import is_file_uri


class ParquetTableCache(BaseTableCache):
    """
    Local Table Cache that stores tables in `Parquet`_ files.

    .. rubric:: Supported Tables

    The `ParquetTableCache` supports Pandas, Polars and pydiverse.transform.


    :param base_path:
        A path to a folder where the Parquet files should get stored.
        To differentiate between different instances, the ``instance_id`` will
        automatically be appended to the provided path.

    .. _parquet:
        https://parquet.apache.org
    """

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        instance_id = normalize_name(ConfigContext.get().instance_id)

        config = config.copy()
        base_path = UPath(config.pop("base_path")) / instance_id
        return cls(base_path=base_path, **config)

    def __init__(self, *args, base_path: str | UPath, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_path = UPath(base_path).absolute()

    def setup(self):
        if is_file_uri(self.base_path):
            os.makedirs(self.base_path, exist_ok=True)

    def init_stage(self, stage: Stage):
        if is_file_uri(self.base_path):
            os.makedirs(self.base_path / stage.name, exist_ok=True)

    def clear_cache(self, stage: Stage):
        _dir = self.get_stage_path(stage)
        if is_file_uri(_dir):
            shutil.rmtree(_dir)
        else:
            _dir.rmdir(recursive=True)

    def _store_table(self, table: Table, task: MaterializingTask):
        if not super()._store_table(table, task):
            return

        metadata = {
            "cache_key": table.cache_key,
        }
        metadata_path = self.get_table_path(table, ".meta.json")
        metadata_path.write_text(json.dumps(metadata))

    def _has_table(self, table: Table, as_type: type) -> bool:
        metadata_path = self.get_table_path(table, ".meta.json")
        if not metadata_path.exists():
            return False

        try:
            metadata = json.loads(metadata_path.read_text(encoding="utf-8"))
            return metadata["cache_key"] == table.cache_key
        except (OSError, json.decoder.JSONDecodeError):
            return False

    def get_stage_path(self, stage: Stage) -> UPath:
        return self.base_path / stage.name

    def get_table_path(self, table: Table, file_extension: str) -> UPath:
        return self.get_stage_path(table.stage) / (table.name + file_extension)


@ParquetTableCache.register_table(pd)
class PandasTableHook(TableHook[ParquetTableCache]):
    pd_version = Version(pd.__version__)

    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, pd.DataFrame))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(type_ == pd.DataFrame)

    @classmethod
    def materialize(
        cls,
        store: ParquetTableCache,
        table: Table,
        stage_name: str,
        without_config_context: bool = False,
    ):
        path = store.get_table_path(table, ".parquet")

        df: pd.DataFrame = table.obj
        df.to_parquet(path)

    @classmethod
    def retrieve(
        cls,
        store: ParquetTableCache,
        table: Table,
        stage_name: str | None,
        as_type: type[pd.DataFrame],
        limit: int | None = None,
    ) -> pd.DataFrame:
        # Determine dtype backend for pandas >= 2.0
        # [this is similar to the PandasTableHook found in SQLTableStore]

        backend_str = "numpy"
        if hook_args := ConfigContext.get().table_hook_args.get("pandas", None):
            if dtype_backend := hook_args.get("dtype_backend", None):
                backend_str = dtype_backend

        if isinstance(as_type, tuple):
            backend_str = as_type[1]
        elif isinstance(as_type, dict):
            backend_str = as_type["backend"]

        if PandasTableHook.pd_version < Version("2.0"):
            # for use_nullable_dtypes=False, returned types are mostly numpy backed
            # extension dtypes
            ret = cls._retrieve(store, table, limit, use_nullable_dtypes=backend_str != "arrow")
            # use_nullable_dtypes=False may still return string[python] even though we
            # expect and sometimes get string[pyarrow]
            for col in ret.dtypes[ret.dtypes == "string[python]"].index:
                ret[col] = ret[col].astype(pd.StringDtype("pyarrow"))
        else:
            dtype_backend_map = {"arrow": "pyarrow", "numpy": "numpy_nullable"}
            ret = cls._retrieve(store, table, limit, dtype_backend=dtype_backend_map[backend_str])

        # Prefer StringDtype("pyarrow") over ArrowDtype(pa.string()) for now.
        # We need to check this choice with future versions of pandas/pyarrow.
        for col in ret.dtypes[(ret.dtypes == "large_string[pyarrow]") | (ret.dtypes == "string[pyarrow]")].index:
            ret[col] = ret[col].astype(pd.StringDtype("pyarrow"))
        return ret

    @classmethod
    def _retrieve(cls, store, table, limit: int | None = None, **pandas_kwargs):
        path = store.get_table_path(table, ".parquet")
        import pyarrow.dataset as ds

        if limit is not None:
            # TODO: pandas_kwargs are currently ignored
            return ds.dataset(path).scanner().head(limit).to_pandas()
        else:
            return pd.read_parquet(path, **pandas_kwargs)


try:
    import polars as pl
except ImportError:
    pl = None


@ParquetTableCache.register_table(pl)
class PolarsTableHook(sql_hooks.PolarsTableHook):
    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, (pl.DataFrame, pl.LazyFrame)))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(issubclass(type_, (pl.DataFrame, pl.LazyFrame)))

    @classmethod
    def materialize(
        cls,
        store: ParquetTableCache,
        table: Table[pl.DataFrame],
        stage_name: str,
        without_config_context: bool = False,
    ):
        path = store.get_table_path(table, ".parquet")
        df = table.obj
        import polars as pl

        if isinstance(df, pl.LazyFrame):
            df = df.collect()
        df.write_parquet(path)
        # intentionally don't apply annotation checks because they might also be done
        # within polars table hook of actual table store

    @classmethod
    def _execute_query(
        cls,
        store: ParquetTableCache,
        table: Table,
        stage_name: str,
        as_type: type,
        dtypes: dict[str, pl.DataType] | None = None,
        limit: int | None = None,
    ) -> pl.DataFrame:
        _ = as_type
        path = store.get_table_path(table, ".parquet")
        df = pl.read_parquet(path, n_rows=limit)
        if issubclass(as_type, pl.LazyFrame):
            return df.lazy()
        return df


try:
    import pydiverse.transform as pdt

    try:
        from pydiverse.transform.eager import PandasTableImpl

        _ = PandasTableImpl

        pdt_old = pdt
        pdt_new = None
    except ImportError:
        try:
            # detect if 0.2 or >0.2 is active
            # this import would only work in <=0.2
            from pydiverse.transform.extended import Polars

            # ensures a "used" state for the import, preventing black from deleting it
            _ = Polars

            pdt_old = None
            pdt_new = pdt
        except ImportError:
            raise NotImplementedError("pydiverse.transform 0.2.0 - 0.2.2 isn't supported") from None
except ImportError:
    pdt = types.ModuleType("pydiverse.transform")
    pdt.Table = None
    pdt_old = None
    pdt_new = None


@ParquetTableCache.register_table(pdt_old)
class PydiverseTransformTableHookOld(TableHook[ParquetTableCache]):
    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        if not issubclass(type_, pdt.Table):
            return CanMatResult.NO
        from pydiverse.transform.eager import PandasTableImpl

        return CanMatResult.YES_BUT_DONT_CACHE if issubclass(type_, PandasTableImpl) else CanMatResult.NO

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        from pydiverse.transform.eager import PandasTableImpl

        return CanRetResult.new(issubclass(type_, PandasTableImpl))

    @classmethod
    def materialize(
        cls,
        store: ParquetTableCache,
        table: Table[pdt.Table],
        stage_name: str,
        without_config_context: bool = False,
    ):
        from pydiverse.transform.core.verbs import collect
        from pydiverse.transform.eager import PandasTableImpl

        t = table.obj
        table = table.copy_without_obj()

        if isinstance(t._impl, PandasTableImpl):
            table.obj = t >> collect()
            return store.get_r_table_hook(pd.DataFrame).materialize(store, table, stage_name)

        raise TypeError(f"Unsupported type {type(t._impl).__name__}")

    @classmethod
    def retrieve(
        cls,
        store: ParquetTableCache,
        table: Table,
        stage_name: str | None,
        as_type: type,
        limit: int | None = None,
    ):
        from pydiverse.transform.eager import PandasTableImpl

        if isinstance(as_type, PandasTableImpl):
            hook = store.get_r_table_hook(pd.DataFrame)
            df = hook.retrieve(store, table, stage_name, pd.DataFrame, limit)
            return pdt.Table(PandasTableImpl(table.name, df))

        raise ValueError(f"Invalid type {as_type}")


@ParquetTableCache.register_table(pdt_new)
class PydiverseTransformTableHook(TableHook[ParquetTableCache]):
    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        if not issubclass(type_, pdt.Table):
            return CanMatResult.NO
        from pydiverse.transform._internal.pipe.verbs import build_query

        query = tbl.obj >> build_query()
        if query is not None:
            # don't cache if the table is SQL backed
            return CanMatResult.NO
        else:
            return CanMatResult.YES_BUT_DONT_CACHE

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        from pydiverse.transform.extended import Polars

        return CanRetResult.new(issubclass(type_, Polars))

    @classmethod
    def materialize(
        cls,
        store: ParquetTableCache,
        table: Table[pdt.Table],
        stage_name: str,
        without_config_context: bool = False,
    ):
        from pydiverse.transform.extended import (
            Polars,
            export,
        )

        t = table.obj
        table = table.copy_without_obj()

        try:
            table.obj = t >> export(Polars(lazy=True))

            hook = store.get_m_table_hook(table)
            return hook.materialize(store, table, stage_name)
        except Exception as e:
            raise TypeError(f"Unsupported type {type(t._ast).__name__}") from e

    @classmethod
    def retrieve(
        cls,
        store: ParquetTableCache,
        table: Table,
        stage_name: str | None,
        as_type: type,
        limit: int | None = None,
    ):
        from pydiverse.transform.extended import Polars

        if isinstance(as_type, Polars):
            import polars as pl

            hook = store.get_r_table_hook(pl.LazyFrame)
            df = hook.retrieve(store, table, stage_name, pd.DataFrame, limit)
            return pdt.Table(df)

        raise ValueError(f"Invalid type {as_type}")


@ParquetTableCache.register_table()
class SqlAlchemyTableHook(TableHook[ParquetTableCache]):
    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        return CanMatResult.NO

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        # retrieving SQLAlchemy reference from parquet table cache does not make sense
        if type_ == sa.Table:
            return CanRetResult.NO_HOOK_IS_EXPECTED
        else:
            return CanRetResult.NO



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from pydiverse.pipedag.backend.lock.zookeeper import KazooClient
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.core.config import PipedagConfig
from tests.fixtures.instances import with_instances
from tests.test_flows.complex_config_flows.test_instance_selection import (
    cfg_file_path,
    check_result,
    get_flow,
)

_ = cfg_file_path
pytestmark = [with_instances("postgres")]


def do_test_lock_manager_instances(cfg_file_path, instance):
    # At this point, an instance is chosen from multi-pipedag-instance
    # configuration file
    pipedag_config = PipedagConfig(cfg_file_path)
    cfg = pipedag_config.get(instance=instance)

    flow, out1, out2 = get_flow(cfg.attrs, pipedag_config)

    with StageLockContext():
        result = flow.run(config=cfg)
        check_result(result, out1, out2)


@pytest.mark.parametrize("instance", ["lock_zookeeper"])
@pytest.mark.skipif(KazooClient is None, reason="requires kazoo")
def test_lock_manager_instances_kazoo(cfg_file_path, instance):
    do_test_lock_manager_instances(cfg_file_path, instance)


@pytest.mark.parametrize("instance", ["lock_file"])
def test_lock_manager_instances(cfg_file_path, instance):
    do_test_lock_manager_instances(cfg_file_path, instance)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import importlib
import pkgutil
from pathlib import Path

import click


@click.group()
def cli():
    pass


def find_commands():
    commands_dir = Path(__file__).parent / "commands"
    return [
        name for _, name, ispkg in pkgutil.iter_modules([str(commands_dir)]) if not ispkg and not name.startswith("_")
    ]


def load_command(command: str):
    importlib.import_module(f"pydiverse.pipedag.management.commands.{command}")


def dynamically_load_commands():
    for command in find_commands():
        load_command(command)


dynamically_load_commands()



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause
import types
from dataclasses import dataclass
from typing import Generic, Mapping, TypeVar

import polars as pl
import pytest
import sqlalchemy as sa
import structlog
from polars.testing import assert_frame_equal

from pydiverse.pipedag import ConfigContext, Flow, Stage, materialize
from pydiverse.pipedag.context.context import CacheValidationMode
from pydiverse.pipedag.errors import HookCheckException
from tests.fixtures.instances import DATABASE_INSTANCES, with_instances

try:
    import dataframely as dy
    from dataframely._polars import FrameType
except ImportError:
    T = TypeVar("T")

    class DyDataFrame(Generic[T]):
        pass

    class DyDummyClass:
        def __init__(self, *args, **kwargs):
            pass

    FrameType = None
    dy = types.ModuleType("dataframely")
    dy.DataFrame = DyDataFrame
    dy.LazyFrame = DyDataFrame
    dy.FailureInfo = None
    dy.Column = None
    dy.Collection = object
    dy.Schema = object
    dy.filter = lambda: lambda fn: fn  # noqa
    dy.rule = lambda: lambda fn: fn  # noqa
    for _type in [
        "Int8",
        "Int16",
        "Int32",
        "Int64",
        "UInt8",
        "UInt16",
        "UInt32",
        "UInt64",
        "Float32",
        "Float64",
        "Bool",
        "String",
        "Decimal",
        "Enum",
        "Struct",
        "List",
        "Date",
        "Datetime",
        "Time",
        "Duration",
        "Float",
        "Integer",
    ]:
        setattr(dy, _type, DyDummyClass)


pytestmark = [
    with_instances(DATABASE_INSTANCES),
]


# ------------------------------------------------------------------------------------ #
#                                        SCHEMA                                        #
# ------------------------------------------------------------------------------------ #


class MyFirstColSpec(dy.Schema):
    a = dy.Integer(primary_key=True)
    b = dy.Int16()
    c = dy.Enum(["x", "y"], nullable=True)


class MySecondColSpec(dy.Schema):
    a = dy.Integer(primary_key=True)
    b = dy.Integer(min=1)
    c = dy.Enum(["x", "y"], nullable=False)


@dataclass
class MyCollection(dy.Collection):
    first: dy.LazyFrame[MyFirstColSpec]
    second: dy.LazyFrame[MySecondColSpec]

    @dy.filter()
    def equal_primary_keys(self) -> pl.LazyFrame:
        return self.first.join(self.second, on=self.common_primary_keys())

    @dy.filter()
    def first_b_greater_second_b(self) -> pl.LazyFrame:
        return self.first.join(self.second, on=self.common_primary_keys(), how="full", coalesce=True).filter(
            (pl.col("b") > pl.col("b_right")).fill_null(True)
        )

    @classmethod
    def _init(cls, data: Mapping[str, FrameType], /):
        return cls(**{k: v.lazy() for k, v in data.items()})


@dataclass
class SimpleCollection(dy.Collection):
    first: dy.LazyFrame[MyFirstColSpec]
    second: dy.LazyFrame[MySecondColSpec]

    @classmethod
    def _init(cls, data: Mapping[str, FrameType], /):
        return cls(**{k: v.lazy() for k, v in data.items()})


# ------------------------------------------------------------------------------------ #
#                                         TESTS                                        #
# ------------------------------------------------------------------------------------ #


enum = pl.Enum(["x", "y"])


def data_without_filter_without_rule_violation() -> tuple[pl.LazyFrame, pl.LazyFrame]:
    first = pl.LazyFrame({"a": [1, 2, 3], "b": [1, 2, 3], "c": ["x", "y", None]}).cast(dict(c=enum))
    second = pl.LazyFrame({"a": [1, 2, 3], "b": [1, 2, 3], "c": ["x", "y", "x"]}).cast(dict(c=enum))
    return first, second


def data_without_filter_with_rule_violation() -> tuple[pl.LazyFrame, pl.LazyFrame]:
    first = pl.LazyFrame({"a": [1, 2, 1], "b": [1, 2, 3], "c": ["x", "y", None]}).cast(dict(c=enum))
    second = pl.LazyFrame({"a": [1, 2, 3], "b": [0, 1, 2], "c": [None, "y", "x"]}).cast(dict(c=enum))
    return first, second


def data_with_filter_without_rule_violation() -> tuple[pl.LazyFrame, pl.LazyFrame]:
    first = pl.LazyFrame({"a": [1, 2, 3], "b": [1, 1, 3], "c": ["x", "y", None]}).cast(dict(c=enum))
    second = pl.LazyFrame({"a": [2, 3, 4, 5], "b": [1, 2, 3, 4], "c": ["x", "y", "x", "x"]}).cast(dict(c=enum))
    return first, second


def data_with_filter_with_rule_violation() -> tuple[pl.LazyFrame, pl.LazyFrame]:
    first = pl.LazyFrame({"a": [1, 2, 3], "b": [1, 2, 3], "c": ["x", "y", None]}).cast(dict(c=enum))
    second = pl.LazyFrame({"a": [2, 3, 4, 5, 6], "b": [0, 1, 2, 3, -1], "c": [None, "y", "y", "x", "z"]})
    return first, second


def test_dataclass():
    first, second = data_without_filter_without_rule_violation()
    c = SimpleCollection(first, second)
    assert_frame_equal(c.first, first)
    assert_frame_equal(c.second, second)


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
def test_enum_violation():
    second = pl.LazyFrame({"a": [2, 3, 4, 5], "b": [0, 1, 2, 3], "c": ["z", "y", "y", "x"]})
    # it is expected that cast fails on invalid enum value
    with pytest.raises(
        pl.exceptions.InvalidOperationError,
        match="conversion from `str` to `enum` failed in column 'c' for 1 out of 4 values",
    ):
        MySecondColSpec.cast(second).collect()
    x, y = MySecondColSpec.filter(second, cast=True)
    assert len(x) == 3
    assert len(y.invalid()) == 1

    class DummyCollection(dy.Collection):
        second: dy.LazyFrame[MySecondColSpec]

    x, y = DummyCollection.filter(dict(second=second), cast=True)
    x = x.second.collect()
    y = y["second"]
    assert len(x) == 3
    assert len(y.invalid()) == 1


@materialize(nout=2)
def get_data(name: str):
    return globals()[f"data_{name}"]()


@materialize(nout=3, input_type=pl.LazyFrame)
def exec_filter_polars(c: dy.Collection):
    out, failure = c.filter(c.__dict__, cast=True)
    return (
        out,
        SimpleCollection(**{name: f._df for name, f in failure.items()}),
        SimpleCollection(**{name: f.counts() for name, f in failure.items()}),
    )


# -------------------------------------- FILTER -------------------------------------- #


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
def test_filter_without_filter_without_rule_violation():
    @materialize(input_type=pl.LazyFrame)
    def assertions(out, failure, failure_counts: dict[str, int]):
        out = out.cast(out.__dict__)
        first, second = data_without_filter_without_rule_violation()

        assert isinstance(out, SimpleCollection)
        assert_frame_equal(out.first, first.cast(dict(b=pl.Int16)))
        assert_frame_equal(out.second, second)
        assert failure.first.select(pl.len()).collect().item() == 0
        assert failure.second.select(pl.len()).collect().item() == 0

    with Flow() as flow:
        with Stage("s01"):
            c = SimpleCollection(*get_data("without_filter_without_rule_violation"))
        with Stage("s02"):
            out, failure, failure_counts = exec_filter_polars(c)
            assertions(out, failure, failure_counts)

    flow.run()


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
def test_filter_without_filter_with_rule_violation():
    @materialize(input_type=pl.LazyFrame)
    def assertions(out, failure, failure_counts: dict[str, int]):
        out = out.cast(out.__dict__)
        # first, second = data_without_filter_with_rule_violation()

        assert isinstance(out, SimpleCollection)
        assert len(out.first.collect()) == 1
        assert len(out.second.collect()) == 2
        assert failure_counts.first == {"primary_key": 2}
        assert failure_counts.second == {"b|min": 1, "c|nullability": 1}

    with Flow() as flow:
        with Stage("s01"):
            c = SimpleCollection(*get_data("without_filter_with_rule_violation"))
        with Stage("s02"):
            out, failure, failure_counts = exec_filter_polars(c)
            assertions(out, failure, failure_counts)

    flow.run()


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
def test_filter_with_filter_without_rule_violation():
    @materialize(input_type=pl.LazyFrame)
    def assertions(out, failure, failure_counts: dict[str, int]):
        out = out.cast(out.__dict__)
        # first, second = data_with_filter_without_rule_violation()

        assert isinstance(out, MyCollection)
        assert_frame_equal(
            out.first,
            pl.LazyFrame({"a": [3], "b": [3], "c": [None]}).cast(dict(b=pl.Int16, c=enum)),
        )
        assert_frame_equal(
            out.second,
            pl.LazyFrame({"a": [3], "b": [2], "c": ["y"]}).cast(dict(c=enum)),
        )
        assert failure_counts.first == {
            "equal_primary_keys": 1,
            "first_b_greater_second_b": 1,
        }
        assert failure_counts.second == {
            "equal_primary_keys": 2,
            "first_b_greater_second_b": 1,
        }

    with Flow() as flow:
        with Stage("s01"):
            c = MyCollection(*get_data("with_filter_without_rule_violation"))
        with Stage("s02"):
            out, failure, failure_counts = exec_filter_polars(c)
            assertions(out, failure, failure_counts)

    flow.run()


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
def test_filter_with_filter_with_rule_violation():
    @materialize(input_type=pl.LazyFrame)
    def assertions(out, failure, failure_counts: dict[str, int]):
        out = out.cast(out.__dict__)
        # first, second = data_with_filter_with_rule_violation()

        assert isinstance(out, MyCollection)
        assert_frame_equal(
            out.first,
            pl.LazyFrame({"a": [3], "b": [3], "c": [None]}).cast(dict(b=pl.Int16, c=enum)),
        )
        assert_frame_equal(
            out.second,
            pl.LazyFrame({"a": [3], "b": [1], "c": ["y"]}).cast(dict(c=enum)),
        )
        assert failure_counts.first == {"equal_primary_keys": 2}
        assert failure_counts.second == {
            "b|min": 1,
            "c|nullability": 1,
            "c|dtype": 1,
            "equal_primary_keys": 2,
        }

    with Flow() as flow:
        with Stage("s01"):
            c = MyCollection(*get_data("with_filter_with_rule_violation"))
        with Stage("s02"):
            out, failure, failure_counts = exec_filter_polars(c)
            assertions(out, failure, failure_counts)

    flow.run()


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
@pytest.mark.parametrize(
    "with_filter, with_violation, validate_get_data",
    [(a, b, c) for a in [False, True] for b in [False, True] for c in [False, True]],
)
def test_annotations(with_filter: bool, with_violation: bool, validate_get_data: bool):
    if validate_get_data:

        @materialize(nout=2)
        def get_anno_data(
            name: str,
        ) -> tuple[dy.LazyFrame[MyFirstColSpec], dy.LazyFrame[MySecondColSpec]]:
            return globals()[f"data_{name}"]()
    else:

        @materialize(nout=2)
        def get_anno_data(name: str) -> tuple[pl.LazyFrame, pl.LazyFrame]:
            return globals()[f"data_{name}"]()

    @materialize(input_type=pl.LazyFrame)
    def consumer(first: dy.LazyFrame[MyFirstColSpec], second: dy.LazyFrame[MySecondColSpec]):
        assert first.collect_schema() == pl.Schema(
            [("a", pl.Int64), ("b", pl.Int16), ("c", pl.Enum(categories=["x", "y"]))]
        )
        assert second.collect_schema() == pl.Schema(
            [("a", pl.Int64), ("b", pl.Int64), ("c", pl.Enum(categories=["x", "y"]))]
        )
        assert len(first.collect()) == 3
        assert len(second.collect()) in [3, 4, 5]

        if not validate_get_data and with_violation:
            with pytest.raises(dy.exc.RuleValidationError, match="1 rules failed validation"):
                MyFirstColSpec.validate(first)
            with pytest.raises(dy.exc.RuleValidationError, match="2 rules failed validation"):
                MySecondColSpec.validate(second)
        else:
            assert MyFirstColSpec.is_valid(first)
            assert MySecondColSpec.is_valid(second)

    @materialize(input_type=dy.LazyFrame)
    def consumer2(first: dy.LazyFrame[MyFirstColSpec], second: dy.LazyFrame[MySecondColSpec]):
        assert first.collect_schema() == pl.Schema(
            [("a", pl.Int64), ("b", pl.Int16), ("c", pl.Enum(categories=["x", "y"]))]
        )
        assert second.collect_schema() == pl.Schema(
            [("a", pl.Int64), ("b", pl.Int64), ("c", pl.Enum(categories=["x", "y"]))]
        )
        assert len(first.collect()) == 3
        assert len(second.collect()) in [3, 4, 5]

    with Flow() as flow:
        name = f"with{'out' if not with_filter else ''}_filter_with{'out' if not with_violation else ''}_rule_violation"
        with Stage("s01"):
            first, second = get_anno_data(name)
            consumer(first, second)
        with Stage("s02"):
            consumer2(first, second)

    if with_violation and validate_get_data:
        # Validation at end of get_anno_data task fails
        with pytest.raises(
            HookCheckException,
            match="failed validation with MyFirstColSpec; Failure counts: "
            "{'b|min': 1, 'c|nullability': 1, 'c|dtype': 1};"
            if with_filter
            else "{'primary_key': 2};",
        ):
            flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    elif with_violation and not validate_get_data and with_filter:
        # Due to the enum failure the dematerialization hook for consumer
        # task fails with ValueError (triggers RuntimeError in Flow)
        with pytest.raises(
            RuntimeError,
            match="Failed to retrieve table '<Table 'get_anno_data",
        ):
            flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    else:
        ret = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
        assert ret.successful


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
@pytest.mark.parametrize(
    "with_filter, with_violation, validate_get_data",
    [(a, b, c) for a in [False, True] for b in [False, True] for c in [False, True]],
)
def test_annotations_not_fail_fast(with_filter: bool, with_violation: bool, validate_get_data: bool):
    if validate_get_data:

        @materialize(nout=2)
        def get_anno_data(
            name: str,
        ) -> tuple[dy.LazyFrame[MyFirstColSpec], dy.LazyFrame[MySecondColSpec]]:
            return globals()[f"data_{name}"]()
    else:

        @materialize(nout=2)
        def get_anno_data(name: str) -> tuple[pl.LazyFrame, pl.LazyFrame]:
            return globals()[f"data_{name}"]()

    @materialize(input_type=pl.LazyFrame)
    def consumer(first: dy.LazyFrame[MyFirstColSpec], second: dy.LazyFrame[MySecondColSpec]):
        assert first.collect_schema() == pl.Schema(
            [("a", pl.Int64), ("b", pl.Int16), ("c", pl.Enum(categories=["x", "y"]))]
        )
        assert second.collect_schema() == pl.Schema(
            [("a", pl.Int64), ("b", pl.Int64), ("c", pl.Enum(categories=["x", "y"]))]
        )
        assert len(first.collect()) == 3
        assert len(second.collect()) in [3, 4, 5]

        assert MyFirstColSpec.is_valid(first)
        assert MySecondColSpec.is_valid(second)

    @materialize(input_type=dy.LazyFrame)
    def consumer2(first: dy.LazyFrame[MyFirstColSpec], second: dy.LazyFrame[MySecondColSpec]):
        assert first.collect_schema() == pl.Schema(
            [("a", pl.Int64), ("b", pl.Int16), ("c", pl.Enum(categories=["x", "y"]))]
        )
        assert second.collect_schema() == pl.Schema(
            [("a", pl.Int64), ("b", pl.Int64), ("c", pl.Enum(categories=["x", "y"]))]
        )
        assert len(first.collect()) == 3
        assert len(second.collect()) in [3, 4, 5]

    with Flow() as flow:
        name = f"with{'out' if not with_filter else ''}_filter_with{'out' if not with_violation else ''}_rule_violation"
        with Stage("s01"):
            first, second = get_anno_data(name)
            consumer(first, second)
        with Stage("s02"):
            consumer2(first, second)

    with ConfigContext.get().evolve(fail_fast=False):
        result = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    if with_violation:
        assert not result.successful
    else:
        assert result.successful


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
@pytest.mark.parametrize(
    "with_filter, with_violation, validate_get_data",
    [(a, b, c) for a in [False, True] for b in [False, True] for c in [False, True]],
)
def test_annotations_fault_tolerant(with_filter: bool, with_violation: bool, validate_get_data: bool):
    if validate_get_data:

        @materialize(nout=2)
        def get_anno_data(
            name: str,
        ) -> tuple[dy.LazyFrame[MyFirstColSpec], dy.LazyFrame[MySecondColSpec]]:
            return globals()[f"data_{name}"]()
    else:

        @materialize(nout=2)
        def get_anno_data(name: str) -> tuple[pl.LazyFrame, pl.LazyFrame]:
            return globals()[f"data_{name}"]()

    @materialize(input_type=pl.LazyFrame)
    def consumer(first: dy.LazyFrame[MyFirstColSpec], second: dy.LazyFrame[MySecondColSpec]):
        assert first.collect_schema() == pl.Schema(
            [("a", pl.Int64), ("b", pl.Int16), ("c", pl.Enum(categories=["x", "y"]))]
        )
        assert second.collect_schema() == pl.Schema(
            [
                ("a", pl.Int64),
                ("b", pl.Int64),
                (
                    "c",
                    pl.Enum(categories=["x", "y"]) if not with_filter or not with_violation else pl.String,
                ),
            ]
        )
        assert len(first.collect()) == 3
        assert len(second.collect()) in [3, 4, 5]

        if with_violation:
            if with_filter:
                MyFirstColSpec.validate(first)
                with pytest.raises(dy.exc.RuleValidationError, match="3 rules failed validation"):
                    MySecondColSpec.validate(second, cast=True)
            else:
                with pytest.raises(dy.exc.RuleValidationError, match="1 rules failed validation"):
                    MyFirstColSpec.validate(first)
                with pytest.raises(dy.exc.RuleValidationError, match="2 rules failed validation"):
                    MySecondColSpec.validate(second)
        else:
            assert MyFirstColSpec.is_valid(first)
            assert MySecondColSpec.is_valid(second)

    @materialize(input_type=dy.LazyFrame)
    def consumer2(first: dy.LazyFrame[MyFirstColSpec], second: dy.LazyFrame[MySecondColSpec]):
        assert len(first.collect()) == 3
        assert len(second.collect()) in [3, 4, 5]

    with Flow() as flow:
        name = f"with{'out' if not with_filter else ''}_filter_with{'out' if not with_violation else ''}_rule_violation"
        with Stage("s01"):
            first, second = get_anno_data(name)
            consumer(first, second)
        with Stage("s02"):
            consumer2(first, second)

    with structlog.testing.capture_logs() as logs:
        with ConfigContext.get().evolve(table_hook_args=dict(polars=dict(fault_tolerant_annotation_action=True))):
            result = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    assert result.successful
    failures = [c for c in logs if c["event"] == "Failed to apply materialize annotation for table"]
    if with_violation and validate_get_data:
        assert len(failures) == 1 if with_filter else 2
        assert all("failed validation with My" in failure["exception"] for failure in failures)


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
@pytest.mark.parametrize(
    "with_filter, with_violation, validate_get_data",
    [(a, b, c) for a in [False, True] for b in [False, True] for c in [False, True]],
)
def test_collections(with_filter: bool, with_violation: bool, validate_get_data: bool):
    CollectionType = MyCollection if with_filter else SimpleCollection

    if validate_get_data:

        @materialize()
        def get_anno_collection(name: str) -> CollectionType:
            first, second = globals()[f"data_{name}"]()
            return CollectionType(first=first, second=second)
    else:

        @materialize()
        def get_anno_collection(name: str):
            first, second = globals()[f"data_{name}"]()
            return CollectionType(first=first, second=second)

    @materialize(input_type=pl.LazyFrame)
    def consumer_collection(coll: CollectionType):
        # # collections are currently not passed on as annotations to individual tables
        # # thus no cast is happening
        # assert coll.first.collect_schema() == pl.Schema([('a', pl.Int64),
        # ('b', pl.Int64), ('c', pl.Enum(categories=['x', 'y']))])
        # assert coll.second.collect_schema() == pl.Schema([('a', pl.Int64),
        # ('b', pl.Int64), ('c', pl.Enum(categories=['x', 'y']))])
        assert len(coll.first.collect()) == 3
        assert len(coll.second.collect()) in [3, 4, 5]

        if with_violation:
            with pytest.raises(dy.exc.MemberValidationError, match="2 members failed validation"):
                coll.validate(coll.__dict__, cast=True)
        else:
            if with_filter:
                # it is not really without violation
                out, _ = coll.filter(coll.__dict__, cast=True)
                assert_frame_equal(
                    out.first,
                    pl.LazyFrame({"a": [3], "b": [3], "c": [None]}).cast(dict(b=pl.Int16, c=enum)),
                )
                assert_frame_equal(
                    out.second,
                    pl.LazyFrame({"a": [3], "b": [2], "c": ["y"]}).cast(dict(c=enum)),
                )
            else:
                assert coll.is_valid(coll.__dict__, cast=True)

    @materialize(input_type=dy.LazyFrame)
    def consumer2_collection(coll: CollectionType):
        # # collections are currently not passed on as annotations to individual tables
        # # thus no cast is happening
        # assert coll.first.collect_schema() == pl.Schema([('a', pl.Int64),
        # ('b', pl.Int64), ('c', pl.Enum(categories=['x', 'y']))])
        # assert coll.second.collect_schema() == pl.Schema([('a', pl.Int64),
        # ('b', pl.Int64), ('c', pl.Enum(categories=['x', 'y']))])
        assert len(coll.first.collect()) == 3
        assert len(coll.second.collect()) in [3, 4, 5]

    with Flow() as flow:
        name = f"with{'out' if not with_filter else ''}_filter_with{'out' if not with_violation else ''}_rule_violation"
        with Stage("s01"):
            collection = get_anno_collection(name)
            consumer_collection(collection)
        with Stage("s02"):
            consumer2_collection(collection)

    # # collections are currently not passed on as annotations to individual tables
    # # thus no cast is happening
    # if with_violation:
    #     from dataframely.exc import RuleValidationError
    #     with pytest.raises(RuleValidationError):
    #         flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    # else:
    ret = flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
    assert ret.successful


@pytest.mark.skipif(dy.Collection is object, reason="dataframely needs to be installed")
def test_type_mapping():
    @materialize(nout=2)
    def get_anno_data() -> tuple[dy.LazyFrame[MyFirstColSpec], dy.LazyFrame[MySecondColSpec]]:
        return data_with_filter_without_rule_violation()

    @materialize(input_type=sa.Table)
    def consumer(first: dy.LazyFrame[MyFirstColSpec], second: dy.LazyFrame[MySecondColSpec]):
        assert isinstance(first.c.b.type, sa.SmallInteger)
        assert isinstance(second.c.b.type, sa.BigInteger)
        assert not isinstance(second.c.b.type, sa.SmallInteger)

    with Flow() as flow:
        with Stage("s01"):
            first, second = get_anno_data()
            consumer(first, second)

    flow.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID)
